"""Module providing video_detection_mscoco functionality."""

import os
import json
import logging
import tempfile
from collections import defaultdict
import cv2
import requests
from uuid import uuid4
from matrice_dataset.server_utils import (
    get_corresponding_split_type,
    generate_short_uuid,
)


def get_video_mscoco_annotations(
    annotation_paths,
):
    """Process MSCOCO-style video dataset annotations and return frame-based annotations grouped
    by split and video, using nested dictionary structure similar to YouTube BB function."""
    complete_videos = {
        "train": {},
        "val": {},
        "test": {},
        "unassigned": {},
    }
    missing_annotations = []
    classwise_stats = {}
    
    for ann_path in annotation_paths:
        if not os.path.exists(ann_path):
            logging.warning(
                "Annotation file not found: %s",
                ann_path,
            )
            continue
            
        filename = os.path.basename(ann_path).lower()
        if "train" in filename:
            split = "train"
        elif "val" in filename:
            split = "val"
        elif "test" in filename:
            split = "test"
        elif "metadata" in filename:
            continue
        else:
            split = "unassigned"
            
        with open(ann_path, "r", encoding="utf-8") as f:
            data = json.load(f)
            
        categories = {cat["id"]: cat["name"] for cat in data.get("categories", [])}
        video_meta = {video["id"]: video for video in data.get("videos", [])}
        
        annotations_by_video = defaultdict(list)
        for ann in data.get("annotations", []):
            annotations_by_video[ann["video_id"]].append(ann)
            
        for video_id, annotations in annotations_by_video.items():
            if video_id not in video_meta:
                logging.warning(
                    "Video ID %s not found in metadata.",
                    video_id,
                )
                continue
                
            video_info = video_meta[video_id]
            video_name = os.path.splitext(video_info["file_name"])[0]
            fps = video_info["fps"]
            video_width = video_info["width"]
            video_height = video_info["height"]
            total_frames = video_info["frames"]
            
            # Create video metadata similar to YouTube BB structure but organized by split
            if video_name not in complete_videos[split]:
                complete_videos[split][video_name] = {
                    "sequence_name": video_name,
                    "splitType": split,
                    "total_frames": total_frames,
                    "frame_rate": fps,
                    "duration_seconds": total_frames / fps,
                    "video_height": video_height,
                    "video_width": video_width,
                    "annotation": {},
                }
            
            # Sort annotations by frame
            annotations = sorted(
                annotations,
                key=lambda x: x["frame_id"],
            )
            
            # Track which frames have annotations
            frames_with_annotations = set()
            
            # Process annotations by frame
            for idx, ann in enumerate(annotations):
                frame_id = ann["frame_id"]
                category_id = ann["category_id"]
                category_name = categories.get(category_id, "unknown")
                
                # Update classwise stats
                if category_name not in classwise_stats:
                    classwise_stats[category_name] = {
                        "train": 0,
                        "val": 0,
                        "test": 0,
                        "unassigned": 0,
                    }
                classwise_stats[category_name][split] += 1
                
                # Calculate time info
                time_start = frame_id / fps
                if idx + 1 < len(annotations):
                    next_frame_id = annotations[idx + 1]["frame_id"]
                else:
                    next_frame_id = video_info["frames"]
                time_end = next_frame_id / fps
                
                # Create annotation similar to MSCOCO but organized by frame
                annotation = {
                    "id": str(generate_short_uuid()),
                    "segmentation": ann.get("segmentation", []),
                    "order_id": frame_id,
                    "file_name": f"{video_name}_{frame_id}.jpg",  # Similar to YouTube BB
                    "file_location": f"{os.path.dirname(ann_path)}/{video_name}/{video_name}_{frame_id}.jpg",  # Similar to YouTube BB
                    "isCrowd": [
                        (float(item) if isinstance(item, (int, float)) else 0)
                        for item in (
                            ann.get("iscrowd", [0])
                            if isinstance(
                                ann.get("iscrowd"),
                                list,
                            )
                            else [ann.get("iscrowd", 0)]
                        )
                    ],
                    "confidence": 0.0,
                    "bbox": ann.get("bbox", []),
                    "height": ann["bbox"][3] if "bbox" in ann and len(ann["bbox"]) > 3 else 0,
                    "width": ann["bbox"][2] if "bbox" in ann and len(ann["bbox"]) > 2 else 0,
                    "center": (
                        [
                            ann["bbox"][0] + ann["bbox"][2] / 2,
                            ann["bbox"][1] + ann["bbox"][3] / 2,
                        ]
                        if ann.get("bbox") and len(ann["bbox"]) > 3
                        else []
                    ),
                    "area": ann.get("area", 0.0),
                    "category": category_name,
                    "masks": [],
                    "visibility": 1.0,  # Similar to YouTube BB's object_presence
                    "duration": [
                        round(time_start, 4),
                        round(time_end, 4),
                    ],  # Keep MSCOCO's time information
                }
                
                # Store the annotation by frame_id as a string key (like YouTube BB)
                frame_key = str(frame_id)
                if frame_key not in complete_videos[split][video_name]["annotation"]:
                    complete_videos[split][video_name]["annotation"][frame_key] = []
                    
                complete_videos[split][video_name]["annotation"][frame_key].append(annotation)
                frames_with_annotations.add(frame_id)
            
            # Track missing annotations
            for frame_id in range(total_frames):
                if frame_id not in frames_with_annotations:
                    missing_file_path = f"{os.path.dirname(ann_path)}/{video_name}/{video_name}_{frame_id}.jpg"
                    missing_annotations.append(missing_file_path)
                    
                    # Add empty annotation list for frames without annotations (like YouTube BB)
                    frame_key = str(frame_id)
                    if frame_key not in complete_videos[split][video_name]["annotation"]:
                        complete_videos[split][video_name]["annotation"][frame_key] = []
    
    # Compute total stats across splits
    for category_name, split_counts in classwise_stats.items():
        split_counts["total"] = sum(v for k, v in split_counts.items() if k != "total")
    
    # Count total videos
    total_videos = sum(len(videos) for split, videos in complete_videos.items())
    
    logging.info(
        "Complete video sequences: %d",
        total_videos,
    )
    logging.info(
        "Frames missing annotations: %d",
        len(missing_annotations),
    )
    logging.info(
        "Class-wise statistics: %s",
        classwise_stats,
    )
    
    return (
        complete_videos,
        missing_annotations,
        classwise_stats,
    )


def get_video_metadata(presigned_url):
    """
    Downloads a video from a presigned URL, extracts its dimensions (width, height),
    FPS, and duration, and saves the first frame as an image locally.

    Args:
        presigned_url (str): The presigned URL of the video.

    Returns:
        dict: {
            "width": video_width,
            "height": video_height,
            "fps": rounded_fps,
            "first_frame_path": path_to_saved_image
        }
    """
    try:
        response = requests.get(presigned_url, stream=True, timeout=30)
        if response.status_code != 200:
            raise Exception(f"Failed to download video: {response.status_code}")
        with tempfile.NamedTemporaryFile(delete=True, suffix=".mp4") as temp_video:
            for chunk in response.iter_content(chunk_size=8192):
                temp_video.write(chunk)
            temp_video.flush()
            cap = cv2.VideoCapture(temp_video.name)
            if not cap.isOpened():
                raise Exception("Failed to open video file")
            width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            fps = cap.get(cv2.CAP_PROP_FPS)
            success, frame = cap.read()
            if not success:
                raise Exception("Failed to read first frame")
            frame_filename = f"first_frame_{uuid4().hex}.jpg"
            cv2.imwrite(frame_filename, frame)
            cap.release()
        rounded_fps = int(fps + 0.5)
        return {
            "width": width,
            "height": height,
            "fps": rounded_fps,
            "first_frame_path": os.path.abspath(frame_filename),
        }
    except Exception as e:
        print(f"Error: {e}")
        return None

def extract_video_identifier(path):
    """Extract the video name from the file path.

    Args:
        path: File path to extract video identifier from

    Returns:
        Video identifier string or None if path is invalid
    """
    if not path:
        return None
    video_folder = os.path.dirname(path)
    video_name = os.path.basename(video_folder)
    return video_name

def add_video_mscoco_dataset_items_details(batch_dataset_items, frames_details):
    """Add MSCOCO video-style annotation details to frame-wise dataset items.

    Args:
        batch_dataset_items: List of dataset items to enhance
        frames_details: Complete videos dictionary from get_video_mscoco_annotations()

    Returns:
        List of processed dataset items with annotation and video metadata
    """
    if not batch_dataset_items or not frames_details:
        return []

    processed_batch = []

    logging.debug(
        "Processing %d dataset items with MSCOCO video annotations",
        len(batch_dataset_items),
    )

    for dataset_item in batch_dataset_items:
        file_info_list = dataset_item.get("fileInfoResponse", [])
        if not isinstance(file_info_list, list) or not file_info_list:
            logging.warning("Missing or invalid fileInfoResponse for dataset item")
            continue

        last_annotations_found = []
        for fileinfo in file_info_list:
            frames_dict = fileinfo.get("frames", {})
            if not isinstance(frames_dict, dict):
                logging.warning("Invalid frames data structure in fileInfoResponse")
                continue

            for frame_key, frame_value in frames_dict.items():
                file_location = frame_value.get("fileLocation")
                if not file_location:
                    logging.warning("Missing fileLocation for frame %s", frame_key)
                    continue

                split = get_corresponding_split_type(file_location)
                video_identifier = extract_video_identifier(file_location)

                if not video_identifier:
                    logging.warning(
                        "Could not extract video identifier from file location: %s",
                        file_location,
                    )
                    continue

                split_data = frames_details.get(split)
                if not split_data:
                    logging.warning("Split '%s' not found in frames_details", split)
                    continue

                video_data = split_data.get(video_identifier)
                if not video_data:
                    logging.warning(
                        "Video identifier '%s' not found in split '%s'",
                        video_identifier,
                        split,
                    )
                    
                    dataset_item.setdefault("splitType", split)
                    dataset_item.setdefault("video_height", 1)
                    dataset_item.setdefault("video_width", 1)
                    frame_value["annotations"] = []

                # Attach video-level metadata
                else:
                    dataset_item.update(
                        {k: v for k, v in video_data.items() if k != "annotation"}
                    )
                    frame_value["annotations"] = video_data.get("annotation", {}).get(frame_key, [])

        processed_batch.append(
            {
                "sample_details": dataset_item,
                "is_complete": all(
                    dataset_item.get(k) is not None
                    for k in ["video_height", "video_width", "splitType"]
                ),
            }
        )
    logging.debug(
        "Successfully processed %d/%d dataset items with MSCOCO annotations",
        len(processed_batch),
        len(batch_dataset_items),
    )
    return processed_batch



#!/usr/bin/env python3

import logging
import json
import sys
from typing import Dict, Any

# Import your pipeline modules
from .image_data_augmentation import (
    create_probability_based_augmentation_pipeline,
    parse_dynamic_pipeline_config,
    DynamicPipelineConfig,
    AugmentationStep,
)

def create_sample_pipeline_config() -> Dict[str, Any]:
    """Create a sample pipeline configuration for testing"""
    # return {
    #     "TotalNewImages": 6,
    #     "augmentationId": "test_aug_001",
    #     "dataset_id": "684916d765b796606a22fc89",
    #     "augChains": [
    #         # Chain 1: Blur + Brightness/Contrast (40 images)
    #         [
    #             [
    #                 {"blur": {"blur_limit": 7, "prob": 1.0}},
    #                 {"brightness_contrast": {"brightness_limit": 0.2, "contrast_limit": 0.2, "prob": 1.0}}
    #             ],
    #             1
    #         ],
    #         # Chain 2: Rotation + Color Jitter (30 images)
    #         [
    #             [
    #                 {"random_affine": {"shift_limit":0.0625, "scale_limit":0.1, "rotate_limit":15, "prob": 1.0}},
    #                 {"color_jitter": {"brightness": 0.2, "contrast": 0.2, "saturation": 0.2, "hue": 0.1, "prob": 1.0}}
    #             ],
    #             1
    #         ],
    #         # Chain 3: Flip + HSV (20 images)
    #         [
    #             [
    #                 {"flip": {"prob": 1.0}},
    #                 {"hsv": {"hue_shift_limit": 20, "sat_shift_limit": 30, "val_shift_limit": 20, "prob": 1.0}}
    #             ],
    #             1
    #         ],
    #         # Chain 4: Noise + Compression (10 images)
    #         [
    #             [
    #                 {"iso_noise": {"color_shift":(0.02, 0.03), "intensity": (0.2, 0.3), "prob": 1.0}},
    #             ],
    #             1
    #         ]
    #     ]
    # }
    
    config_data = [
        
        {
        "TotalNewImages": 15,
        "augmentationId": "aug_123",
        "dataset_id": "6800f1ab79ee9ad0a1bc8749",
        "minAugmentationsPerImage": 2,
        "maxAugmentationsPerImage": 6,
        "augmentationPool": [
            
                    {"blur": {"blur_limit": 7}},
                    {"brightness_contrast": {"brightness_limit": 0.2, "contrast_limit": 0.2}},
                    {"random_affine": {"shift_limit": 0.0625, "scale_limit": 0.1, "rotate_limit": 15}},
                    {"color_jitter": {"brightness": 0.2, "contrast": 0.2, "saturation": 0.2, "hue": 0.1}},
                    {"flip": {}},  # No parameters needed for flip
                    {"hsv": {"hue_shift_limit": 20, "sat_shift_limit": 30, "val_shift_limit": 20}}
            ]
        
    },
                   ]
    return config_data

def create_kafka_config() -> Dict[str, Any]:
    """Create Kafka configuration with hardcoded values"""
    return {
        # Kafka broker configuration
        "bootstrap_servers": ["35.238.188.103:9092"],
        
        "Kafka_ip": "",
        "Kafka_port": "",
        # Topics for dataset pagination
        "dataset_request_topic": "augmentation_dataset_item_request",
        "dataset_response_topic": "augmentation_dataset_item_response",
        
        # Topics for upload URL management
        "upload_url_request_topic": "augmentation_PreSignedURL_request",
        "upload_url_response_topic": "augmentation_PreSignedURL_response",
        
        # Topic for publishing new augmented dataset items
        "new_items_topic": "update_augmented_dataset_items",
        
        # Consumer groups
        "consumer_group": "augmentation_pipeline_consumer",
        "upload_url_consumer_group": "upload_url_consumer",
        
        # Pagination settings - conservative to prevent over-consumption
        "page_size": 5  # Small page size to prevent over-queuing
    }

def create_completion_api_config() -> str:
    """Create completion API URL configuration"""
    # Use environment variable if available, otherwise default to None to skip API calls
    import os
    return os.getenv('COMPLETION_API_URL', None)  # Returns None if not set, which will skip API calls
    
    

def initialize_and_run_pipeline():
    """Initialize and run the probability-based augmentation pipeline"""
    
    try:
        logging.info("=== Starting Probability-Based Augmentation Pipeline ===")
        
        # Create sample configuration
        config_data = create_sample_pipeline_config()
        logging.info(f"Created pipeline configuration: {json.dumps(config_data, indent=2)}")
        
        # Parse pipeline configuration
        pipeline_config = parse_dynamic_pipeline_config(config_data)
        logging.info(f"Parsed pipeline config - Total images: {pipeline_config.total_new_images}")
        logging.info(f"Augmentation ID: {pipeline_config.augmentation_id}")
        logging.info(f"Dataset ID: {pipeline_config.dataset_id}")
        # logging.info(f"Number of augmentation chains: {len(pipeline_config.augmentation_chains)}")
        
        # Log chain details
        # for i, chain in enumerate(pipeline_config.augmentation_chains):
        #     logging.info(f"Chain {i}: {len(chain.steps)} steps, target: {chain.target_count} images")
        #     for j, step in enumerate(chain.steps):
        #         logging.info(f"  Step {j}: {step.name} with params {step.params}")
        
        # Create Kafka configuration
        kafka_config = create_kafka_config()
        logging.info(f"Kafka configuration: {json.dumps(kafka_config, indent=2)}")
        
        # Completion API URL (configurable)
        completion_api_url = create_completion_api_config()
        if completion_api_url:
            logging.info(f"Using completion API URL: {completion_api_url}")
        else:
            logging.info("No completion API URL configured - will skip API calls")
        
        # Create the pipeline
        logging.info("Creating probability-based augmentation pipeline...")
        pipeline = create_probability_based_augmentation_pipeline(
            pipeline_config=pipeline_config,
            kafka_config=kafka_config,
        )
        
        if pipeline is None:
            logging.error("Failed to create pipeline")
            return False
        
        logging.info("Pipeline created successfully!")
        logging.info("Pipeline stages:")
        for i, stage in enumerate(pipeline.stages):
            logging.info(f"  Stage {i}: {stage}")
        
        # Start the pipeline
        logging.info("Starting pipeline execution...")
        pipeline.start()
        
        logging.info("=== Pipeline execution completed successfully ===")
        return True
        
    except KeyboardInterrupt:
        logging.info("Pipeline execution interrupted by user")
        return False
    except Exception as e:
        logging.error(f"Error during pipeline execution: {e}")
        import traceback
        traceback.print_exc()
        return False

def run_with_custom_config(config_file_path: str = None):
    """Run pipeline with custom configuration from file"""
    
    try:
        if config_file_path:
            logging.info(f"Loading configuration from: {config_file_path}")
            with open(config_file_path, 'r') as f:
                config_data = json.load(f)
        else:
            logging.info("Using default sample configuration")
            config_data = create_sample_pipeline_config()
        
        # Parse and run
        pipeline_config = parse_dynamic_pipeline_config(config_data)
        kafka_config = create_kafka_config()
        completion_api_url = create_completion_api_config()
        
        if completion_api_url:
            logging.info(f"Using completion API URL: {completion_api_url}")
        else:
            logging.info("No completion API URL configured - will skip API calls")
        
        # Create and start pipeline
        pipeline = create_probability_based_augmentation_pipeline(
            pipeline_config=pipeline_config,
            kafka_config=kafka_config,
        )
        
        if pipeline:
            logging.info("Starting pipeline execution...")
            pipeline.start()
            logging.info("Pipeline execution completed successfully")
            return True
        else:
            logging.error("Failed to create pipeline")
            return False
            
    except Exception as e:
        logging.error(f"Error running pipeline with custom config: {e}")
        import traceback
        traceback.print_exc()
        return False

def transform_augmentation_data(input_dict):
    """
    Transform input augmentation dictionary into a list of output dictionaries,
    one for each augmentation chain.
    """
    def map_operation_to_pool_format(operation):
        """Convert operation format to augmentationPool format"""
        name = operation["Name"]
        params = operation.get("Params", {})
        return {name: params}

    def clean_augmentation_id(aug_id):
        """Clean ObjectID string to extract just the ID"""
        if isinstance(aug_id, str) and aug_id.startswith('ObjectID("') and aug_id.endswith('")'):
            return aug_id[10:-2]
        return aug_id

    result = []

    for chain in input_dict.get("augChains", []):
        augmentation_pool = [map_operation_to_pool_format(op) for op in chain.get("Operations", [])]

        result.append({
            "TotalNewImages": chain.get("NumImagesToGenerate", 0),
            "augmentationId": clean_augmentation_id(input_dict.get("augmentationID", "unknown")),
            "dataset_id": input_dict.get("dataset_id", ""),
            "minAugmentationsPerImage": 1,
            "maxAugmentationsPerImage": len(augmentation_pool),
            "augmentationPool": augmentation_pool
        })

    return result

        
class DataAugmentation:
    """Class to handle dataset preparation."""

    def __init__(self, session: Any, action_record_id: str):
        """Initialize DataAugmentation.

        Args:
            session: Session object with RPC client
            action_record_id: ID of action record
        """
        self.session = session
        self.rpc = session.rpc
        self.action_record_id = action_record_id

        # Updated to correctly parse nested "data"
        url = f"/v1/project/action/{self.action_record_id}/details"
        response = self.rpc.get(url)
        self.action_doc = response["data"]

        # Extract the appropriate fields from the nested structure
        self.action_type = self.action_doc.get("action")
        self.job_params = self.action_doc.get("jobParams", {})
        self.action_details = self.action_doc.get("actionDetails", {})

        logging.info(f"Job parameters: {self.job_params}")

        # Now extract nested fields correctly
        self.dataset_id = self.action_details.get("_idDataset")
        self.source_dataset_version = self.job_params.get("source_version", "")
        self.target_dataset_version = self.job_params.get("target_version", "")
        self.augmentations = self.job_params.get("augmentation_dict", {})

        # Transform using updated function
        self.formatted_augmentations = transform_augmentation_data(self.augmentations)

        self.kafka_config = self.job_params.get("kafka_config", {})
        # self.completion_api_url = self.job_params.get("completion_api_url", "")


    def update_status(
        self,
        stepCode: str,
        status: str,
        status_description: str,
    ) -> None:
        """Update status of data preparation.

        Args:
            stepCode: Code indicating current step
            status: Status of step
            status_description: Description of status
        """
        try:
            logging.info(status_description)
            url = "/v1/actions"
            payload = {
                "_id": self.action_record_id,
                "action": self.action_type,
                "serviceName": self.action_doc["serviceName"],
                "stepCode": stepCode,
                "status": status,
                "statusDescription": status_description,
            }
            self.rpc.put(path=url, payload=payload)
        except Exception as e:
            logging.error(
                "Exception in update_status: %s",
                str(e),
            )
            raise

    def start_processing(self) -> None:
        """Start dataset augmentation processing."""
        self.update_status(
            "DCKR_PROC",
            "OK",
            "Dataset started augmentation",
        )

        for chain in self.formatted_augmentations:
            try:
                logging.info(f"Processing augmentation chain: {json.dumps(chain, indent=2)}")
                
                self.pipeline_config = parse_dynamic_pipeline_config(
                    chain, self.source_dataset_version, self.target_dataset_version
                )
                logging.debug(f"Parsed pipeline config: {self.pipeline_config}")
                for attr, value in vars(self.pipeline_config).items():
                    logging.info(f"Pipeline Config Attribute - {attr}: {value}")
                
                self.pipeline = create_probability_based_augmentation_pipeline(
                    pipeline_config=self.pipeline_config,
                    kafka_config=self.kafka_config,
                )
                self.pipeline.start()
                self.pipeline.wait_to_finish_processing_and_stop()
                new_dataset_items = self.pipeline.get_all_items_from_last_stage()

                # Optional: Uncomment these if you want to preserve partition updates
                # update_partitions_numbers(
                #     self.rpc,
                #     self.dataset_id,
                #     new_dataset_items,
                #     partition_key="partitionNumber",
                # )

                # create_partition_stats(
                #     rpc=self.rpc,
                #     dataset_id=self.dataset_id,
                #     source_version=self.dataset_version,
                #     target_version=self.dataset_version,
                #     partition_stats=calculate_partition_stats(new_dataset_items),
                # )

                # batch_insert_dataset_items(
                #     new_dataset_items,
                #     self.dataset_id,
                #     self.rpc,
            except Exception as e:
                logging.error(f"Error processing augmentation chain: {e}")
                continue    # )

        try:
            completion_url = f"/v1/dataset_item/complete/{self.pipeline_config.augmentation_id}"
            logging.debug(f"Calling completion API at {completion_url}")
            response = self.rpc.post(path=completion_url, payload={})

            if response and response.get("success"):
                logging.info("Completion API called successfully")
            else:
                logging.warning(f"Completion API call returned status {response}")
        except Exception as e:
            logging.error(f"Error calling completion API: {e}")

              # Continue to next chain even if this one fails

        self.update_status(
            "SUCCESS",
            "SUCCESS",
            "Dataset Augmentation completed",
        )



"""Module providing augmentation server functionality."""

import logging
import threading
import urllib.request
from typing import List, Dict, Any, Optional
import signal
import atexit
import base64
import io

import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from PIL import Image
import cv2  # Import OpenCV
import httpx # Use httpx for async uploads

from matrice_common.utils import dependencies_check

# Add opencv-python-headless to the dependency check
dependencies_check(["albumentations", "httpx", "opencv-python-headless"])

from matrice_dataset.image_augmentations import (  # noqa: E42
    get_augmentation_compose,
)


class AugmentationRequest(BaseModel):
    """Request model for augmentation endpoint."""

    image_url: Optional[str] = None
    image_base64: Optional[str] = None
    # ADDED: A pre-signed URL from the backend to upload the result to
    upload_url: str
    image_bboxes: List[Dict] = []
    augmentation_configs: Dict[str, Dict[str, Any]] = {}


class AugmentationResponse(BaseModel):
    """Response model for augmentation endpoint."""

    success: bool
    message: str
    # CHANGED: This now returns the URL of the uploaded image, not the base64 data
    augmented_image_url: Optional[str] = None
    augmented_bboxes: Optional[List[Dict]] = None


class AugmentationServer:
    """Class to handle dataset augmentation server."""

    def __init__(
        self, session: Any, action_record_id: str, port: int, ip_address: str = None
    ):
        """Initialize AugmentationServer.

        Args:
            session: Session object with RPC client
            action_record_id: ID of action record
            port: Port to run the server on
            ip_address: IP address to bind to (optional)
        """

        self.session = session
        self.rpc = session.rpc
        self.action_record_id = action_record_id
        self.port = port
        self.ip = ip_address or self._get_external_ip()
        self._shutdown_event = threading.Event()
        self._server_thread = None

        # Initialize FastAPI app
        self.app = FastAPI(title="Matrice Augmentation Server", version="1.0.0")
        self._setup_routes()
        self._setup_shutdown_handlers()

        # Fetch action details
        self._fetch_action_details()

        logging.info(f"Job parameters: {self.job_params}")
        logging.info("Successfully initialized augmentation server on IP: %s", self.ip)

    def _get_external_ip(self) -> str:
        """Get external IP address."""
        try:
            response = urllib.request.urlopen("https://ident.me", timeout=60)
            return response.read().decode("utf8").strip()
        except Exception as e:
            logging.warning(f"Failed to get external IP: {e}, using localhost")
            return "localhost"

    def _fetch_action_details(self) -> None:
        """Fetch action details from the API."""
        try:
            url = f"/v1/project/action/{self.action_record_id}/details"
            response = self.rpc.get(url)
            self.action_doc = response["data"]

            # Extract the appropriate fields from the nested structure
            self.action_type = self.action_doc.get("action")
            self.job_params = self.action_doc.get("jobParams", {})
            self.action_details = self.action_doc.get("actionDetails", {})
            self.augmentation_server_id = self.action_details.get(
                "augmentation_server_id",
                self.job_params.get("augmentation_server_id", None),
            )
        except Exception as e:
            logging.error(f"Failed to fetch action details: {e}")
            # Set defaults
            self.action_doc = {}
            self.action_type = "augmentation"
            self.job_params = {}
            self.action_details = {}
            self.augmentation_server_id = None

    def _setup_routes(self) -> None:
        """Setup FastAPI routes."""

        @self.app.post("/augment", response_model=AugmentationResponse)
        async def augment_dataset_item(request: AugmentationRequest):
            """Augment dataset item and upload to a pre-signed URL."""
            try:
                # Validate input
                if not request.image_url and not request.image_base64:
                    raise HTTPException(
                        status_code=400,
                        detail="Either image_url or image_base64 must be provided",
                    )
                if not request.upload_url:
                    raise HTTPException(
                        status_code=400, detail="An upload_url must be provided"
                    )
                if not request.image_bboxes:
                    raise HTTPException(status_code=400, detail="No bboxes provided")
                if not request.augmentation_configs:
                    raise HTTPException(
                        status_code=400, detail="No augmentation configs provided"
                    )

                # Load image from URL or Base64
                image = await self._load_image(request.image_url, request.image_base64)

                # Apply augmentations
                augmentation_fn = get_augmentation_compose(request.augmentation_configs)
                augmented_result = augmentation_fn(
                    image=image,
                    bboxes=request.image_bboxes,
                    class_labels=[None for _ in range(len(request.image_bboxes))],
                )

                # --- NEW UPLOAD LOGIC ---
                augmented_image_np = augmented_result["image"]

                # Convert RGB numpy array (from albumentations) to BGR for OpenCV
                image_bgr = cv2.cvtColor(augmented_image_np, cv2.COLOR_RGB2BGR)

                # Encode image to JPEG format in memory
                is_success, img_encoded = cv2.imencode(".jpg", image_bgr)
                if not is_success:
                    raise HTTPException(
                        status_code=500, detail="Failed to encode augmented image."
                    )
                
                img_bytes = img_encoded.tobytes()

                # Upload the image bytes to the pre-signed S3 URL using httpx
                async with httpx.AsyncClient() as client:
                    upload_response = await client.put(
                        request.upload_url,
                        content=img_bytes,
                        headers={'Content-Type': 'image/jpeg'},
                        timeout=60.0,
                    )

                # Check for a successful upload
                if upload_response.status_code not in [200, 201, 204]:
                    error_detail = f"Failed to upload image. Cloud storage responded with status: {upload_response.status_code}. Response: {upload_response.text}"
                    logging.error(error_detail)
                    raise HTTPException(status_code=500, detail=error_detail)

                # Get the permanent URL (the part of the URL before the '?' query string)
                permanent_url = request.upload_url.split('?')[0]
                # --- END OF NEW LOGIC ---

                return AugmentationResponse(
                    success=True,
                    message="Augmentation completed and image uploaded successfully",
                    augmented_image_url=permanent_url,  # Return the permanent URL
                    augmented_bboxes=augmented_result.get("bboxes", []),
                )

            except HTTPException:
                raise
            except Exception as e:
                logging.error(f"Error in augmentation: {str(e)}", exc_info=True)
                raise HTTPException(
                    status_code=500, detail=f"Internal server error: {str(e)}"
                )

        @self.app.get("/health")
        async def health_check():
            """Health check endpoint."""
            return {"status": "healthy", "server": "augmentation_server"}

        @self.app.get("/")
        async def root():
            """Root endpoint."""
            return {"message": "Matrice Augmentation Server", "version": "1.0.0"}

    async def _load_image(
        self, image_url: Optional[str], image_base64: Optional[str]
    ) -> Any:
        """Load image from URL or base64 string."""
        try:
            if image_url:
                # Handle both local file paths and URLs
                if image_url.startswith(("http://", "https://")):
                    async with httpx.AsyncClient() as client:
                        response = await client.get(image_url)
                        response.raise_for_status()
                        image = Image.open(io.BytesIO(response.content))
                else:
                    # Local file path
                    image = Image.open(image_url)
            else:
                # Base64 image
                image_data = base64.b64decode(image_base64)
                image = Image.open(io.BytesIO(image_data))

            # Convert to RGB if necessary
            if image.mode != "RGB":
                image = image.convert("RGB")

            return image

        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f"Failed to load image: {str(e)}"
            )

    def _image_to_base64(self, image: Any) -> str:
        """Convert PIL Image or NumPy array to base64 string."""
        try:
            # Convert numpy array to PIL Image if necessary
            if hasattr(image, "shape"):  # numpy array
                import numpy as np

                if image.dtype != np.uint8:
                    image = (image * 255).astype(np.uint8)
                image = Image.fromarray(image)

            # Convert to base64
            buffer = io.BytesIO()
            image.save(buffer, format="JPEG")
            image_base64 = base64.b64encode(buffer.getvalue()).decode()
            return image_base64

        except Exception as e:
            raise HTTPException(
                status_code=500, detail=f"Failed to convert image to base64: {str(e)}"
            )
            
    # ... The rest of the file (setup_shutdown_handlers, update_status, etc.) remains the same ...
    def _setup_shutdown_handlers(self) -> None:
        """Setup shutdown signal handlers."""

        def signal_handler(signum, frame):
            """Handle shutdown signals."""
            logging.info(f"Received signal {signum}, shutting down gracefully...")
            self.stop_server()

        def atexit_handler():
            """Handle atexit cleanup."""
            try:
                self.stop_server()
            except Exception as exc:
                logging.error("Error during atexit shutdown: %s", str(exc))

        # Register signal handlers
        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

        # Register atexit handler
        atexit.register(atexit_handler)

        logging.info("Shutdown handlers registered successfully")

    def update_status(
        self,
        stepCode: str,
        status: str,
        status_description: str,
    ) -> None:
        """Update status of augmentation server.

        Args:
            stepCode: Code indicating current step
            status: Status of step
            status_description: Description of status
        """
        try:
            logging.info(status_description)
            url = "/v1/actions"
            payload = {
                "_id": self.action_record_id,
                "action": self.action_type,
                "serviceName": self.action_doc["serviceName"],
                "stepCode": stepCode,
                "status": status,
                "statusDescription": status_description,
            }
            self.rpc.put(path=url, payload=payload)
        except Exception as e:
            logging.error(
                "Exception in update_status: %s",
                str(e),
            )
            # Don't raise here to avoid breaking the server

    def update_server_address(self) -> None:
        """Update server address in the backend."""
        try:
            path = "/v1/actions/augmentation_servers"
            payload = {
                "action_id": self.action_record_id,
                "ip_address": self.ip,
                "port": self.port,
                "augmentation_server_id": self.augmentation_server_id,
            }
            resp = self.rpc.put(path=path, payload=payload)
            logging.info(f"Server address update response: {resp}")
        except Exception as e:
            logging.error(f"Failed to update server address: {e}")

    def start_server(self) -> None:
        """Start the augmentation server."""
        try:
            self.update_status(
                "DCKR_PROC",
                "OK",
                "Starting augmentation server",
            )

            # Update server address in backend
            self.update_server_address()

            # Start the server in a background thread
            def run_server():
                """Run the uvicorn server."""
                try:
                    logging.info(
                        "Starting augmentation server on %s:%d",
                        self.ip,
                        self.port,
                    )
                    server = uvicorn.Server(
                        uvicorn.Config(
                            app=self.app,
                            host="0.0.0.0",
                            port=self.port,
                            log_level="info",
                        )
                    )
                    server.run()
                except Exception as exc:
                    logging.error(
                        "Failed to start augmentation server: %s",
                        str(exc),
                    )
                    self.update_status(
                        "ERROR",
                        "ERROR",
                        f"Failed to start server: {str(exc)}",
                    )
                    raise

            # Start the server in a background thread
            self._server_thread = threading.Thread(
                target=run_server, daemon=False, name="AugmentationServer"
            )
            self._server_thread.start()

            # Wait a moment to ensure server started
            import time

            time.sleep(2)

            self.update_status(
                "SUCCESS",
                "SUCCESS",
                f"Augmentation server started successfully on {self.ip}:{self.port}",
            )

            logging.info("Augmentation server thread started successfully")

        except Exception as e:
            logging.error(f"Failed to start server: {e}")
            self.update_status(
                "ERROR",
                "ERROR",
                f"Failed to start augmentation server: {str(e)}",
            )
            raise

    def stop_server(self) -> None:
        """Stop the augmentation server gracefully."""
        if not self._shutdown_event.is_set():
            logging.info("Stopping augmentation server...")
            self._shutdown_event.set()

            # Update status
            try:
                self.update_status(
                    "STOPPED",
                    "STOPPED",
                    "Augmentation server stopped",
                )
            except Exception as e:
                logging.error(f"Failed to update stop status: {e}")

    def wait_for_shutdown(self) -> None:
        """Wait for the server to be shut down."""
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join()

# Flow
# 1. Start the server by creating action and adding the aggmentaion server in the DB
# 2. Once the server with the action starts it will use the update address api call to save its ip and port
# 3. The client will make request to augment an image
# 4. We will get the address and port from the DB
# 5. Make request to the server "http://ip:port/augment" and get the augmented image


"""Module providing data_augmentation functionality."""

import os
import cv2
import logging
from typing import Any, List, Dict
from queue import Queue
from matrice_dataset.pipeline import (
    Pipeline,
)
from matrice_dataset.data_prep import (
    dataset_items_producer,
)
from matrice_dataset.client import (
    add_batch_presigned_upload_urls,
    upload_batch_files,
)
from matrice_dataset.server_utils import (
    download_file,
    generate_short_uuid,
)
from matrice_dataset.client_utils import (
    update_partitions_numbers,
    create_partition_stats,
    SAMPLES_PARTITION_TYPE,
)
from matrice_dataset.image_augmentations import (
    get_augmentation_compose,
)


def create_augmentation_fns(
    augmentation_configs: Dict,
) -> List[callable]:
    """Create a function to perform data augmentation.

    Args:
        augmentation_configs: Dictionary containing augmentation parameters

    Returns:
        List of augmentation functions
    """
    augmentations = []
    for augmentation_config in augmentation_configs:
        augmentations.append(get_augmentation_compose(augmentation_config))
    return augmentations


def load_image(dataset_item):
    image = cv2.imread(dataset_item["filename"])
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def load_bboxes(dataset_item):
    bboxes = []
    for info in dataset_item["versionInfo"]:
        for annotation in info["annotations"]:
            bboxes.append(annotation["bbox"])
    return bboxes


def augment_dataset_items(batch_dataset_items, augmentation_fns):
    augmented_dataset_items = []
    for dataset_item in batch_dataset_items:
        if not dataset_item:
            continue
        for augmentation_fn in augmentation_fns:
            augmented_dataset_item = dataset_item.copy()
            image = load_image(dataset_item)
            bboxes = load_bboxes(dataset_item)
            augmented_dataset_item["augmentation_result"] = augmentation_fn(
                image=image,
                bboxes=bboxes,
                class_labels=[None for _ in range(len(bboxes))],
            )
            augmented_dataset_items.append(augmented_dataset_item)
    return augmented_dataset_items


def save_augmented_images(batch_dataset_items, dataset_id):
    for item in batch_dataset_items:
        new_filename = f"{generate_short_uuid()}.{os.path.splitext(item['filename'])[-1]}"
        augmentation_result = item["augmentation_result"]
        cv2.imwrite(
            new_filename,
            augmentation_result["image"],
        )
        item["filename"] = new_filename
        item["local_file_path"] = new_filename
        item["cloud_file_path"] = f"{dataset_id}/{new_filename}"
        for i in range(len(augmentation_result["bboxes"])):
            item["versionInfo"][0]["annotation"][i]["bbox"] = augmentation_result["bboxes"][i]
    return batch_dataset_items


def filter_dataset_items(
    batch_dataset_items: Dict[str, List],
    dataset_version: str = "v1.0",
) -> List[Dict]:
    """Filter dataset items based on the set type (train/test/val).

    Args:
        dataset_items: Dictionary containing version info and items
        dataset_version: Version string to filter by

    Returns:
        List of training items for the specified version
    """
    filtered_dataset_items = []
    for item in batch_dataset_items:
        for info in item["versionInfo"]:
            if info["version"] == dataset_version and info["itemSetType"] == "train":
                item["versionInfo"] = [info]
                filtered_dataset_items.append(item)
    return filtered_dataset_items


def download_images(
    dataset_items: List[Dict],
) -> List[Dict]:
    """Download images for dataset items.

    Args:
        dataset_items: List of dataset items containing file locations and names

    Returns:
        List of successfully downloaded items
    """
    downloaded_images = []
    for dataset_item in dataset_items:
        try:
            download_file(
                dataset_item["fileLocation"],
                dataset_item["filename"],
            )
            downloaded_images.append(dataset_item)
        except Exception as e:
            logging.error(
                "Error downloading image %s: %s",
                dataset_item["filename"],
                str(e),
            )
    return downloaded_images


def batch_insert_dataset_items(batch, dataset_id, rpc):
    logging.debug("batch to insert: %s", batch)
    payload = {
        "datasetId": dataset_id,
        "datasetItems": batch,
    }
    resp = rpc.post(
        "/v2/dataset/insert_dataset_items",
        payload=payload,
    )
    logging.info("Response from insert: %s", resp)
    return batch


def calculate_partition_stats(
    dataset_items: List[Dict],
) -> List[Dict]:
    """Calculate partition statistics for dataset items.

    Args:
        dataset_items: List of dataset items
    """
    partition_stats = {}
    for item in dataset_items:
        if item["partitionNumber"] not in partition_stats:
            partition_stats[item["partitionNumber"]] = {
                "partitionNum": item["partitionNumber"],
                "sampleCount": 0,
                "diskSizeMB": 0,
                "type": SAMPLES_PARTITION_TYPE,
            }
        partition_stats[item["partitionNumber"]]["sampleCount"] += 1
        partition_stats[item["partitionNumber"]]["diskSizeMB"] += os.path.getsize(
            item["local_file_path"]
        ) / (1024 * 1024)
    return list(partition_stats.values())


def get_data_augmentation_pipeline(
    rpc: Any,
    dataset_id: str,
    dataset_version: str,
    augmentation_configs: list,
    max_attempts: int = 5,
    bucket_alias: str = "",
    account_number: str = "",
) -> Pipeline:
    """Get the data augmentation pipeline.

    Args:
        rpc: RPC client for making API calls
        dataset_id: Dataset ID
        dataset_version: Dataset version
        augmentation_configs: List of augmentation configurations
        max_attempts: Maximum number of upload retry attempts
        bucket_alias: Storage bucket alias
        account_number: Account number for storage access

    Returns:
        Configured Pipeline object for data augmentation
    """
    dataset_items_queue = Queue()
    download_images_queue = Queue()
    dataset_items_augmentation_queue = Queue()
    update_dataset_items_queue = Queue()
    save_augmented_images_queue = Queue()
    upload_files_queue = Queue()
    insert_dataset_items_queue = Queue()
    pipeline = Pipeline()
    pipeline.add_producer(
        process_fn=dataset_items_producer,
        process_params={
            "rpc": rpc,
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "pipeline_queue": dataset_items_queue,
        },
    )
    pipeline.add_stage(
        stage_name="Filter Dataset Items",
        process_fn=filter_dataset_items,
        pull_queue=dataset_items_queue,
        push_queue=download_images_queue,
        process_params={"dataset_version": dataset_version},
        num_threads=1,
    )
    pipeline.add_stage(
        stage_name="Download Original Images",
        process_fn=download_images,
        pull_queue=download_images_queue,
        push_queue=dataset_items_augmentation_queue,
        num_threads=15,
    )
    pipeline.add_stage(
        stage_name="Augment Dataset Items",
        process_fn=augment_dataset_items,
        pull_queue=dataset_items_augmentation_queue,
        push_queue=update_dataset_items_queue,
        process_params={"augmentation_fns": create_augmentation_fns(augmentation_configs)},
        num_threads=10,
    )
    pipeline.add_stage(
        stage_name="Fetch Upload Presigned URLs",
        pull_queue=update_dataset_items_queue,
        push_queue=save_augmented_images_queue,
        process_fn=add_batch_presigned_upload_urls,
        process_params={
            "rpc": rpc,
            "partition_type": SAMPLES_PARTITION_TYPE,
            "bucket_alias": bucket_alias,
            "account_number": account_number,
        },
        num_threads=10,
    )
    pipeline.add_stage(
        stage_name="Save Augmented Images",
        pull_queue=save_augmented_images_queue,
        push_queue=upload_files_queue,
        process_fn=save_augmented_images,
        process_params={"dataset_id": dataset_id},
        num_threads=10,
    )
    pipeline.add_stage(
        stage_name="Upload Augmented Images",
        pull_queue=upload_files_queue,
        push_queue=insert_dataset_items_queue,
        process_fn=upload_batch_files,
        process_params={"max_attempts": max_attempts},
        num_threads=10,
    )
    return pipeline


class DataAugmentation:
    """Class to handle dataset preparation."""

    def __init__(self, session: Any, action_record_id: str):
        """Initialize DataAugmentation.

        Args:
            session: Session object with RPC client
            action_record_id: ID of action record
        """
        self.session = session
        self.rpc = session.rpc
        self.action_record_id = action_record_id
        url = f"/v1/project/action/{self.action_record_id}/details"
        self.action_doc = self.rpc.get(url)["data"]
        self.action_type = self.action_doc["action"]
        self.job_params = self.action_doc["jobParams"]
        self.dataset_id = self.job_params["dataset_id"]
        self.dataset_version = self.job_params["dataset_version"]
        self.augmentations = self.job_params["augmentations"]

    def update_status(
        self,
        stepCode: str,
        status: str,
        status_description: str,
    ) -> None:
        """Update status of data preparation.

        Args:
            stepCode: Code indicating current step
            status: Status of step
            status_description: Description of status
        """
        try:
            logging.info(status_description)
            url = "/v1/actions"
            payload = {
                "_id": self.action_record_id,
                "action": self.action_type,
                "serviceName": self.action_doc["serviceName"],
                "stepCode": stepCode,
                "status": status,
                "statusDescription": status_description,
            }
            self.rpc.put(path=url, payload=payload)
        except Exception as e:
            logging.error(
                "Exception in update_status: %s",
                str(e),
            )
            raise

    def start_processing(self) -> None:
        """Start dataset augmentation processing."""
        self.update_status(
            "DCKR_PROC",
            "OK",
            "Dataset started augmentation",
        )
        self.pipeline = get_data_augmentation_pipeline(
            rpc=self.rpc,
            dataset_id=self.dataset_id,
            dataset_version=self.dataset_version,
            augmentation_configs=self.augmentations,
        )
        self.pipeline.start()
        self.pipeline.wait_to_finish_processing_and_stop()
        new_dataset_items = self.pipeline.get_all_items_from_last_stage()
        update_partitions_numbers(
            self.rpc,
            self.dataset_id,
            new_dataset_items,
            partition_key="partitionNumber",
        )
        create_partition_stats(
            rpc=self.rpc,
            dataset_id=self.dataset_id,
            source_version=self.dataset_version,
            target_version=self.dataset_version,
            partition_stats=calculate_partition_stats(new_dataset_items),
        )
        batch_insert_dataset_items(
            new_dataset_items,
            self.dataset_id,
            self.rpc,
        )
        self.update_status(
            "SUCCESS",
            "SUCCESS",
            "Dataset Augmentation completed",
        )



"""Module providing labelbox_detection functionality."""

import os
import re
import json
import logging
import traceback
import shutil
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse
import requests
from matrice_dataset.server_utils import (
    generate_short_uuid,
    download_file,
)


def load_ndjson(file_path):
    """Reads an NDJSON file and extracts valid JSON objects.

    Args:
        file_path: Path to the NDJSON file

    Returns:
        List of parsed JSON objects
    """
    with open(file_path, "r", encoding="utf-8") as f:
        raw_text = f.read()
    json_objects = re.findall(
        "\\{.*?\\}(?=\\s*\\{|\\s*$)",
        raw_text,
        re.DOTALL,
    )
    parsed_data = []
    for obj in json_objects:
        try:
            parsed_data.append(json.loads(obj))
        except json.JSONDecodeError as e:
            logging.error("Invalid JSON entry: %s", e)
    return parsed_data


def calculate_bbox_properties(bbox):
    """Calculate properties for a bounding box.

    Args:
        bbox: Bounding box in format [x_min, y_min, width, height]

    Returns:
        Dictionary of properties including height, width, center, area

    Raises:
        ValueError: If bbox doesn't have exactly 4 elements
    """
    if len(bbox) != 4:
        raise ValueError("Bounding box must be in the format [x_min, y_min, width, height].")
    x_min, y_min, width, height = bbox
    center_x = x_min + width / 2
    center_y = y_min + height / 2
    area = width * height
    return {
        "height": float(height),
        "width": float(width),
        "center": [
            float(center_x),
            float(center_y),
        ],
        "area": float(area),
    }


def get_labelbox_image_details(annotation_files):
    """Process Labelbox NDJSON annotation files and extract image details.

    Args:
        annotation_files (list): List of paths to Labelbox NDJSON annotation files.

    Returns:
        tuple:
        - Dictionary of image details indexed by file location.
        - List of image entries missing annotations.
        - Dictionary of class-wise splits.
    """
    complete_images = {}
    missing_annotations = []
    missing_dimensions = {}
    classwise_splits = {}
    logging.info(
        "Processing %s annotation files",
        len(annotation_files),
    )
    if not annotation_files:
        logging.error("No annotation files provided")
        return (
            complete_images,
            missing_annotations,
            classwise_splits,
        )
    for file_index, file_path in enumerate(annotation_files, 1):
        logging.debug(
            "\nProcessing file %s/%s: %s",
            file_index,
            len(annotation_files),
            file_path,
        )
        if not os.path.exists(file_path):
            logging.error("File not found: %s", file_path)
            continue
        entries = load_ndjson(file_path)
        for entry in entries:
            try:
                row_data = entry.get("data_row", {}).get("row_data", "")
                parsed_url = urlparse(row_data)
                split_type = "unassigned"
                for segment in parsed_url.path.split("/"):
                    if segment in [
                        "train",
                        "val",
                        "test",
                    ]:
                        split_type = segment
                        break
                external_id = entry["data_row"].get("external_id", "unknown")
                media_attrs = entry.get("media_attributes", {})
                height = media_attrs.get("height", None)
                width = media_attrs.get("width", None)
                annotations = []
                for project in entry.get("projects", {}).values():
                    for label in project.get("labels", []):
                        for obj in label.get("annotations", {}).get("objects", []):
                            category = obj.get("name")
                            bbox = obj.get("bounding_box", {})
                            if not category or not bbox:
                                continue
                            if category not in classwise_splits:
                                classwise_splits[category] = {
                                    "train": 0,
                                    "val": 0,
                                    "test": 0,
                                    "unassigned": 0,
                                }
                            classwise_splits[category][split_type] += 1
                            bbox_entry = [
                                float(bbox.get("left", 0)),
                                float(bbox.get("top", 0)),
                                float(bbox.get("width", 0)),
                                float(
                                    bbox.get(
                                        "height",
                                        0,
                                    )
                                ),
                            ]
                            bbox_properties = calculate_bbox_properties(bbox_entry)
                            annotation = {
                                "id": str(generate_short_uuid()),
                                "segmentation": [],
                                "isCrowd": [],
                                "confidence": 0.0,
                                "bbox": bbox_entry,
                                "height": bbox_properties["height"],
                                "width": bbox_properties["width"],
                                "center": bbox_properties["center"],
                                "area": bbox_properties["area"],
                                "category": str(category),
                                "masks": [],
                            }
                            annotations.append(annotation)
                key = f"{external_id}"
                details = {
                    "splitType": split_type,
                    "file_name": external_id,
                    "image_url": row_data,
                    "annotations": annotations,
                }
                if height is not None and width is not None:
                    details.update(
                        {
                            "image_height": height,
                            "image_width": width,
                            "image_area": height * width,
                        }
                    )
                else:
                    missing_dimensions[key] = details
                if key in complete_images:
                    complete_images[key]["annotations"].extend(annotations)
                else:
                    complete_images[key] = details
            except Exception as e:
                logging.error(
                    "Error processing entry: %s",
                    e,
                )
                traceback.print_exc()
    for (
        category,
        counts,
    ) in classwise_splits.items():
        counts["total"] = sum(counts.values())
    logging.info("\nFinal summary:")
    logging.info(
        "Complete images: %s",
        len(complete_images),
    )
    logging.info(
        "Missing annotations: %s",
        len(missing_annotations),
    )
    logging.info(
        "Missing dimensions: %s",
        len(missing_dimensions),
    )
    return (
        {**complete_images, **missing_dimensions},
        missing_annotations,
        classwise_splits,
    )


def add_labelbox_dataset_items_details(batch_dataset_items, images_details):
    """Add image details to batch dataset items.

    Args:
        batch_dataset_items: List of dataset items to process
        images_details: Dictionary of image details indexed by file name

    Returns:
        List of processed dataset items with details added
    """
    processed_batch = []
    for dataset_item in batch_dataset_items:
        image_key = f"{dataset_item.get('filename')}"
        if image_key not in images_details:
            logging.warning(
                "'%s' not found in images_details",
                image_key,
            )
            continue
        dataset_item.update(images_details[image_key])
        processed_batch.append(
            {
                "sample_details": dataset_item,
                "is_complete": all(
                    k in images_details[image_key]
                    for k in [
                        "image_height",
                        "image_width",
                    ]
                ),
            }
        )
    return processed_batch


def add_labelbox_dataset_item_local_file_path(batch_dataset_items, base_dataset_path):
    """Add local file paths to dataset items.

    Args:
        batch_dataset_items: List of dataset items to process
        base_dataset_path: Base path for dataset

    Returns:
        List of processed dataset items with local file paths added
    """
    processed_batch = []
    for dataset_item in batch_dataset_items:
        if not dataset_item.get("is_complete"):
            image_path = f"{base_dataset_path}/images/{dataset_item.get('filename')}"
            if not os.path.exists(image_path):
                logging.warning(
                    "Image not found, will download it: %s",
                    image_path,
                )
                download_file(
                    dataset_item["sample_details"]["image_url"],
                    image_path,
                )
            dataset_item["sample_details"]["local_file_path"] = image_path
        processed_batch.append(dataset_item)
    return processed_batch


def download_labelbox_images(images_path, labelbox_annotations_path):
    """
    Download Labelbox images from the given annotation path and save them to the images path.
    Uses multithreading to download images in parallel.

    Args:
        images_path: Path to save the downloaded images
        labelbox_annotations_path: Path to the labelbox annotations file
    """
    os.makedirs(images_path, exist_ok=True)
    annotations = load_ndjson(labelbox_annotations_path)
    
    def download_single_image(entry):
        try:
            data_row = entry.get("data_row", {})
            image_url = data_row.get("row_data")
            image_name = data_row.get("external_id")
            if not image_url or not image_name:
                return
            
            image_path = os.path.join(images_path, image_name)
            if os.path.exists(image_path):
                return
            
            try:
                response = requests.get(
                    image_url,
                    stream=True,
                    timeout=30,
                )
                response.raise_for_status()
                with open(image_path, "wb") as img_file:
                    for chunk in response.iter_content(chunk_size=8192):
                        img_file.write(chunk)
                logging.debug(
                    "Downloaded %s successfully.",
                    image_name,
                )
            except requests.exceptions.RequestException as e:
                logging.error(
                    "Failed to download %s: %s",
                    image_name,
                    e,
                )
        except Exception as e:
            logging.error("Error processing entry: %s", e)
    
    # Use ThreadPoolExecutor to parallelize downloads
    with ThreadPoolExecutor(max_workers=min(10, len(annotations))) as executor:
        # Submit all download tasks
        list(executor.map(download_single_image, annotations))


def download_labelbox_dataset_items(dataset_id, labelbox_annotations_path):
    """
    Download Labelbox dataset from the given annotation path and save it to the base dataset path

    Args:
        dataset_id: ID for the dataset
        labelbox_annotations_path: Path to the labelbox annotations file

    Returns:
        Dataset ID
    """
    annotations_dir = f"{dataset_id}/annotations/"
    images_dir = f"{dataset_id}/images"
    os.makedirs(images_dir, exist_ok=True)
    os.makedirs(annotations_dir, exist_ok=True)
    shutil.copy2(
        labelbox_annotations_path,
        os.path.join(
            annotations_dir,
            os.path.basename(labelbox_annotations_path),
        ),
    )
    logging.debug(
        "Copied annotation file from %s to %s",
        labelbox_annotations_path,
        annotations_dir,
    )
    download_labelbox_images(images_dir, labelbox_annotations_path)
    return dataset_id



"""Module providing pascalvoc_detection functionality."""

import os
import xml.etree.ElementTree as ET
import logging
from matrice_dataset.server_utils import (
    generate_short_uuid,
)


def calculate_pascal_bbox_properties(bbox):
    """Calculate properties for Pascal VOC bounding box.

    Args:
        bbox: Bounding box in format [x_min, y_min, x_max, y_max]

    Returns:
        Dictionary of properties including bbox, height, width, center, area

    Raises:
        ValueError: If bbox doesn't have exactly 4 elements
    """
    if len(bbox) != 4:
        raise ValueError("Bounding box must be in the format [x_min, y_min, x_max, y_max].")
    x_min, y_min, x_max, y_max = bbox
    width = x_max - x_min
    height = y_max - y_min
    center_x = x_min + width / 2
    center_y = y_min + height / 2
    area = width * height
    return {
        "bbox": [
            float(x_min),
            float(y_min),
            float(width),
            float(height),
        ],
        "height": float(height),
        "width": float(width),
        "center": [
            float(center_x),
            float(center_y),
        ],
        "area": float(area),
    }


def get_pascalvoc_image_details(annotation_files):
    """Process Pascal VOC annotation files and extract image details.

    Args:
        annotation_files: List of paths to Pascal VOC annotation files (XML and TXT)

    Returns:
        Tuple containing:
        - Dictionary of image details indexed by file location
        - List of image filenames missing annotations
        - Dictionary of class-wise splits
    """
    complete_images = {}
    missing_annotations = []
    missing_dimensions = {}
    classwise_splits = {}
    logging.info(
        "Processing %s annotation files",
        len(annotation_files),
    )
    if not annotation_files:
        logging.warning("No annotation files provided")
        return (
            complete_images,
            missing_annotations,
            classwise_splits,
        )
    xml_files = [f for f in annotation_files if f.lower().endswith(".xml")]
    txt_files = [
        f
        for f in annotation_files
        if f.lower().endswith(".txt") and "ImageSets/Main" in f.replace("\\", "/")
    ]
    logging.debug(
        "Processing %s XML files: %s",
        len(xml_files),
        xml_files,
    )
    logging.debug(
        "Processing %s TXT files: %s",
        len(txt_files),
        txt_files,
    )
    split_mapping = {}
    for txt_file in txt_files:
        split_type = os.path.splitext(os.path.basename(txt_file))[0]
        if split_type in ["train", "val", "test"]:
            if not os.path.exists(txt_file):
                logging.error("File not found: %s", txt_file)
                continue
            with open(txt_file, "r", encoding="utf-8") as f:
                for line in f:
                    image_name = line.strip()
                    split_mapping[image_name] = split_type
    logging.info(
        "Processing %s XML files to extract annotations",
        len(xml_files),
    )
    for xml_file in xml_files:
        if not os.path.exists(xml_file):
            logging.error("File not found: %s", xml_file)
            continue
        try:
            tree = ET.parse(xml_file)
            root = tree.getroot()
            filename = root.find("filename").text
            filename_without_extension = os.path.splitext(filename)[0]
            split_type = split_mapping.get(
                filename_without_extension,
                "unassigned",
            )
            size = root.find("size")
            try:
                width = int(size.find("width").text)
                height = int(size.find("height").text)
            except (TypeError, ValueError):
                missing_dimensions[filename] = {
                    "file_name": filename,
                    "splitType": split_type,
                }
                width, height = None, None
            annotations = []
            for obj in root.findall("object"):
                category = obj.find("name").text
                bndbox = obj.find("bndbox")
                bbox = [
                    float(bndbox.find("xmin").text),
                    float(bndbox.find("ymin").text),
                    float(bndbox.find("xmax").text),
                    float(bndbox.find("ymax").text),
                ]
                bbox_properties = calculate_pascal_bbox_properties(bbox)
                annotation_json = {
                    "id": str(generate_short_uuid()),
                    "segmentation": [],
                    "isCrowd": [],
                    "confidence": 0.0,
                    "bbox": bbox_properties["bbox"],
                    "height": bbox_properties["height"],
                    "width": bbox_properties["width"],
                    "center": bbox_properties["center"],
                    "area": bbox_properties["area"],
                    "category": str(category),
                    "masks": [],
                }
                annotations.append(annotation_json)
                if category not in classwise_splits:
                    classwise_splits[category] = {
                        "train": 0,
                        "val": 0,
                        "test": 0,
                        "unassigned": 0,
                    }
                classwise_splits[category][split_type] += 1
            key = f"{filename}"
            details = {
                "splitType": split_type,
                "file_name": filename,
                "image_height": height,
                "image_width": width,
                "image_area": (width * height if width and height else None),
                "annotations": annotations,
            }
            if annotations:
                complete_images[key] = details
            else:
                missing_annotations.append(filename)
        except ET.ParseError as e:
            logging.error(
                "Error parsing XML %s: %s",
                xml_file,
                e,
            )
        except Exception as e:
            logging.error(
                "Error processing file %s: %s",
                xml_file,
                e,
            )
    logging.info("\nFinal summary:")
    logging.info(
        "Complete images: %s",
        len(complete_images),
    )
    logging.info(
        "Missing annotations: %s",
        len(missing_annotations),
    )
    logging.info(
        "Missing dimensions: %s",
        len(missing_dimensions),
    )
    for (
        category,
        counts,
    ) in classwise_splits.items():
        counts["total"] = sum(counts.values())
    return (
        {**complete_images, **missing_dimensions},
        missing_annotations,
        classwise_splits,
    )


def add_pascalvoc_dataset_items_details(batch_dataset_items, images_details):
    """Add image details to batch dataset items.

    Args:
        batch_dataset_items: List of dataset items to process
        images_details: Dictionary of image details indexed by filename

    Returns:
        List of processed dataset items with details added
    """
    processed_batch = []
    for dataset_item in batch_dataset_items:
        image_key = f"{dataset_item.get('filename')}"
        if image_key not in images_details:
            logging.warning(
                "'%s' not found in images_details",
                image_key,
            )
            continue
        dataset_item.update(images_details[image_key])
        processed_batch.append(
            {
                "sample_details": dataset_item,
                "is_complete": all(
                    dataset_item.get(k) is not None
                    for k in [
                        "image_height",
                        "image_width",
                    ]
                ),
            }
        )
    return processed_batch



"""Module providing client functionality."""

import os
from queue import Queue
import logging
from matrice_dataset.pipeline import Pipeline
from concurrent.futures import ThreadPoolExecutor, as_completed
from matrice_dataset.client_utils import (
    ANNOTATION_PARTITION_TYPE,
    SAMPLES_PARTITION_TYPE,
    scan_dataset,
    get_annotations_partition,
    get_images_partitions,
    get_youtube_bb_partitions,
    get_cloud_file_path,
    update_annotation_bucket_url,
    get_batch_pre_signed_upload_urls,
    upload_file,
    compress_annotation_files,
    update_partitions_numbers,
    create_partition_stats,
    get_youtube_bb_relative_path,
    get_mot_partitions,
    get_video_mot_cloud_file_path,
    get_video_mscoco_cloud_file_path,
    get_davis_partitions,
    get_davis_relative_path,
    get_video_imagenet_partitions,
    get_kinetics_partitions,
    get_video_mscoco_partitions,
    create_video_blank_dataset_items,
    extract_frames_from_videos,
    restructure_davis_dataset
)


def get_partition_status(base_path, skip_annotation_partition=False):
    logging.info("Getting partition status")
    annotation_partition, images_partitions = (
        None,
        [],
    )
    annotation_files, image_files = scan_dataset(base_path)
    if annotation_files:
        logging.debug("Debug - annotation_files length: %s", len(annotation_files))
    if image_files:
        logging.debug("Debug - image_files length: %s", len(image_files))
    if not skip_annotation_partition:
        annotation_partition = get_annotations_partition(annotation_files)
    images_partitions = get_images_partitions(image_files)
    if annotation_partition:
        logging.debug("Debug - annotation_partition length: %s", len(annotation_partition))
    else:
        logging.debug("Debug - annotation_partition is None")
    if images_partitions:
        logging.debug("Debug - images_partitions length: %s", len(images_partitions))
    else:
        logging.debug("Debug - images_partitions is None")
    return annotation_partition, images_partitions


def get_video_partition_status(
    base_path,
    skip_annotation_partition=False,
    get_partitions=get_youtube_bb_partitions,
    rename_annotation_files=False,
    input_type=None,
):
    logging.info(
        "Getting partition %s status with base path %s",
        input_type,
        base_path,
    )
    annotation_partition, images_partitions = (
        None,
        [],
    )
    annotation_files, image_files = scan_dataset(
        base_path,
        rename_annotation_files,
        input_type,
    )
    if not skip_annotation_partition:
        annotation_partition = get_annotations_partition(annotation_files)
    unique_videos=None
    images_partitions, unique_videos = get_partitions(image_files)
    return annotation_partition, images_partitions, unique_videos


def get_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    files = partition["files"]

    def create_file_info(file_path):
        file_name = os.path.basename(file_path)
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": file_name,
            "cloud_file_path": get_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    if len(files) <= batch_size:
        return [[create_file_info(f) for f in files]]
    batches = []
    num_batches = (len(files) + batch_size - 1) // batch_size
    for i in range(num_batches):
        start_idx = i * batch_size
        end_idx = min(start_idx + batch_size, len(files))
        batch_files = files[start_idx:end_idx]
        batches.append([create_file_info(f) for f in batch_files])
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    return batches


def get_video_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]

    def create_file_info(file_path, transformed_frame_number):
        file_name = os.path.basename(file_path)
        video_identifier, frame_number = file_name.rsplit("_", 1)
        frame_number = int(frame_number.split(".")[0])
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": file_name,
            "video_identifier": video_identifier,
            "frame_number": frame_number,
            "transformed_frame_number": transformed_frame_number,
            "cloud_file_path": get_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []
    for video_files in videos:
        if not video_files:
            continue
        video_info = [create_file_info(file, idx) for idx, file in enumerate(video_files)]
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "batches from get_video_partition_batches: %s",
        batches,
    )
    return batches


def get_video_mot_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]
    logging.debug("batch_size is %s", batch_size)

    def create_file_info(file_path):
        os.path.basename(os.path.dirname(file_path))
        video_identifier = os.path.basename(os.path.dirname(os.path.dirname(file_path)))
        path_parts = file_path.split(os.sep)[::-1]
        split = "unlabeled"
        for part in path_parts:
            if part.lower() in {
                "train",
                "test",
                "val",
            }:
                split = part
                break
        file_name = os.path.basename(file_path)
        frame_number = int(os.path.splitext(file_name)[0])
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": file_name,
            "video_identifier": video_identifier,
            "frame_number": frame_number,
            "cloud_file_path": get_video_mot_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
            "split": split,
        }

    batches = []
    current_batch = []
    for video_files in videos:
        if not video_files:
            continue
        video_info = [create_file_info(file) for file in video_files]
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "batches from get_video_partition_batches: %s",
        batches,
    )
    return batches


def get_video_davis_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]
    partition_type = partition["type"]  # train, val, or test
    logging.debug("batch_size is %s", batch_size)

    def create_file_info(file_path):
        video_dir = os.path.dirname(file_path)                     # .../partition_type/video_name
        video_name = os.path.basename(video_dir)                  # video_1
        partition_dir = os.path.basename(os.path.dirname(video_dir))  # train / val / test
        video_identifier = f"{partition_dir}_{video_name}"        # e.g., train_video_1
        file_name = os.path.basename(file_path)
        frame_number = int(os.path.splitext(file_name)[0])
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition_type,
            "local_file_path": file_path,
            "file_name": file_name,
            "video_identifier": video_identifier,
            "frame_number": frame_number,
            "cloud_file_path": get_video_mot_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []
    for video_files in videos:
        if not video_files:
            continue
        video_info = [create_file_info(file) for file in video_files]
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)

    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "batches from get_video_partition_batches: %s",
        batches,
    )
    return batches


def get_video_imagenet_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]
    logging.debug("Batch size is %s", batch_size)
    logging.debug(
        "In-get_video_imagenet_partition_batches-videos are %s",
        videos,
    )

    def create_file_info(file_path):
        video_name = os.path.basename(file_path)
        class_folder = os.path.basename(os.path.dirname(file_path))
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": video_name,
            "video_identifier": class_folder,
            "cloud_file_path": get_video_mot_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []
    for video_file in videos:
        if not video_file:
            continue
        video_info = create_file_info(video_file)
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "Batches from get_video_davis_partition_batches: %s",
        batches,
    )
    return batches


def get_video_kinetics_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]
    logging.debug("Batch size is %s", batch_size)
    logging.debug(
        "In-get_video_kinetics_partition_batches-videos are %s",
        videos,
    )

    def create_file_info(file_path):
        video_name = os.path.basename(file_path)
        class_folder = os.path.basename(os.path.dirname(file_path))
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": video_name,
            "video_identifier": class_folder,
            "cloud_file_path": get_video_mot_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []
    for video_file in videos:
        if not video_file:
            continue
        video_info = create_file_info(video_file)
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "Batches from get_video_davis_partition_batches: %s",
        batches,
    )
    return batches


def get_video_mscoco_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]
    partition_type = partition["type"]  # train, val, test
    logging.debug("Batch size is %s", batch_size)

    def create_file_info(file_path):
        file_name = os.path.basename(file_path)                   # e.g., 000111.jpg
        video_dir = os.path.dirname(file_path)                    # .../test/video_003
        video_name = os.path.basename(video_dir)                  # video_003
        partition_dir = os.path.basename(os.path.dirname(video_dir))  # test
        video_identifier = f"{partition_dir}/{video_name}"        # e.g., test/video_003
        frame_number = int(os.path.splitext(file_name)[0])        # e.g., 111

        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition_type,
            "local_file_path": file_path,
            "file_name": file_name,
            "video_identifier": video_identifier,
            "frame_number": frame_number,
            "cloud_file_path": get_video_mscoco_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []

    for video_file in videos:
        if not video_file:
            continue
        video_info = [create_file_info(video_path) for video_path in video_file]
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)

    if current_batch:
        batches.append(current_batch)

    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "Batches from get_video_mscoco_partition_batches: %s",
        batches,
    )
    return batches



def get_video_youtube_bb_partition_batches(
    partition,
    batch_size,
    dataset_id,
    dataset_version,
    base_dataset_path,
    include_version_in_cloud_path=False,
):
    videos = partition["files"]

    def create_file_info(file_path, transformed_frame_number):
        file_name = os.path.basename(file_path)
        video_identifier, frame_number = file_name.rsplit("_", 1)
        frame_number = int(frame_number.split(".")[0])
        return {
            "partition_num": partition["partitionNum"],
            "partition_type": partition["type"],
            "local_file_path": file_path,
            "file_name": file_name,
            "video_identifier": video_identifier,
            "frame_number": frame_number,
            "transformed_frame_number": transformed_frame_number,
            "cloud_file_path": get_video_mot_cloud_file_path(
                dataset_id,
                dataset_version,
                base_dataset_path,
                file_path,
                include_version_in_cloud_path,
            ),
        }

    batches = []
    current_batch = []
    for video_files in videos:
        if not video_files:
            continue
        video_info = [create_file_info(file, idx) for idx, file in enumerate(video_files)]
        if len(current_batch) >= batch_size:
            batches.append(current_batch)
            current_batch = []
        current_batch.append(video_info)
    if current_batch:
        batches.append(current_batch)
    logging.info(
        "Created %s batches for partition %s",
        len(batches),
        partition["partitionNum"],
    )
    logging.debug(
        "batches from get_video_partition_batches: %s",
        batches,
    )
    return batches


def add_batch_presigned_upload_urls(
    batch,
    rpc,
    partition_type,
    bucket_alias="",
    account_number="",
    project_id=None,
):
    logging.debug("batch to add presigned urls: %s", batch)
    cloud_paths_presigned_url_dict = get_batch_pre_signed_upload_urls(
        [file_info["cloud_file_path"] for file_info in batch],
        rpc,
        partition_type,
        bucket_alias,
        account_number,
        project_id=project_id,
    )
    for file_info in batch:
        file_info["presigned_url"] = cloud_paths_presigned_url_dict.get(
            file_info["cloud_file_path"], None
        )
    return batch


def add_video_batch_presigned_upload_urls(
    batch,
    rpc,
    partition_type,
    bucket_alias="",
    account_number="",
):
    frames = [file_info for video in batch for file_info in video]
    logging.debug("batch to add presigned urls: %s", frames)
    cloud_paths_presigned_url_dict = get_batch_pre_signed_upload_urls(
        [file_info["cloud_file_path"] for file_info in frames],
        rpc,
        partition_type,
        bucket_alias,
        account_number,
    )
    for file_info in frames:
        file_info["presigned_url"] = cloud_paths_presigned_url_dict.get(
            file_info["cloud_file_path"], None
        )
    logging.debug(
        "final batch for add_video_batch_presigned_upload_urls: %s",
        batch,
    )
    return batch


def add_video_imagenet_presigned_upload_urls(
    batch,
    rpc,
    partition_type,
    bucket_alias="",
    account_number="",
):
    videos = [video for video in batch]
    logging.debug("batch to add presigned urls: %s", videos)
    cloud_paths_presigned_url_dict = get_batch_pre_signed_upload_urls(
        [file_info["cloud_file_path"] for file_info in videos],
        rpc,
        partition_type,
        bucket_alias,
        account_number,
    )
    for file_info in videos:
        file_info["presigned_url"] = cloud_paths_presigned_url_dict.get(
            file_info["cloud_file_path"], None
        )
    logging.debug(
        "final batch for add_video_batch_presigned_upload_urls: %s",
        batch,
    )
    return batch


def upload_batch_files(batch, max_attempts=5):
    for file_info in batch:
        success = upload_file(
            file_info["local_file_path"],
            file_info["presigned_url"],
            max_attempts,
        )
        file_info["upload_success"] = success
    return batch


def upload_video_batch_files(batch, max_attempts=5):
    logging.debug(
        "upload_video_batch_files with max_attempts %s",
        max_attempts,
    )
    # Flatten the batch in one line
    frames = [file_info for video in batch for file_info in video]
    
    # Process all files in the batch using multithreading
    def upload_single_file(file_info):
        success = upload_file(
            file_info["local_file_path"],
            file_info["presigned_url"],
            max_attempts,
        )
        file_info["upload_success"] = success
        return file_info
    
    # Use ThreadPoolExecutor to parallelize uploads
    with ThreadPoolExecutor(max_workers=min(10, len(frames))) as executor:
        # Submit all upload tasks
        future_to_file = {executor.submit(upload_single_file, file_info): file_info for file_info in frames}
        
        # Process results as they complete
        for future in as_completed(future_to_file):
            try:
                future.result()
            except Exception as exc:
                file_info = future_to_file[future]
                logging.error(f"File upload failed for {file_info['local_file_path']}: {exc}")
                file_info["upload_success"] = False
        
    logging.debug(
        "final batch for add_video_batch_presigned_upload_urls: %s",
        batch,
    )
    return batch


def upload_video_imagenet_batch_files(batch, max_attempts=5):
    logging.debug(
        "batch received in upload_video_imagenet_batch_files: %s",
        batch,
    )
    logging.debug(
        "upload_video_batch_files with max_attempts %s",
        max_attempts,
    )
    
    # Use ThreadPoolExecutor to parallelize uploads
    with ThreadPoolExecutor(max_workers=min(10, len(batch))) as executor:
        # Define upload function for a single file
        def upload_single_file(file_info):
            success = upload_file(
                file_info["local_file_path"],
                file_info["presigned_url"],
                max_attempts,
            )
            file_info["upload_success"] = success
            return file_info
        
        # Submit all upload tasks
        future_to_file = {executor.submit(upload_single_file, file_info): file_info for file_info in batch}
        
        # Process results as they complete
        for future in as_completed(future_to_file):
            try:
                future.result()
            except Exception as exc:
                file_info = future_to_file[future]
                logging.error(f"File upload failed for {file_info['local_file_path']}: {exc}")
                file_info["upload_success"] = False
    
    logging.debug(
        "final batch for add_video_batch_presigned_upload_urls: %s",
        batch,
    )
    return batch


def upload_video_mot_batch_files(batch, max_attempts=5):
    frames = [file_info for video in batch for file_info in video]
    for file_info in frames:
        success = upload_file(
            file_info["cloud_file_path"],
            file_info["presigned_url"],
            max_attempts,
        )
        file_info["upload_success"] = success
    logging.debug(
        "final batch for add_video_batch_presigned_upload_urls: %s",
        batch,
    )
    return batch


def batch_create_dataset_items(batch, dataset_id, dataset_version, rpc):
    logging.debug("batch to create dataset items: %s", batch)
    payload = {
        "datasetId": dataset_id,
        "partitionNumber": batch[0]["partition_num"],
        "version": dataset_version,
        "files": [
            {
                "fileName": item["file_name"],
                "fileLocation": item["cloud_file_path"],
            }
            for item in batch
        ],
    }
    resp = rpc.post(
        "/v1/dataset_item/add_dataset_items",
        payload=payload,
    )
    logging.info(
        "Response from create dataset items: %s",
        resp,
    )
    return batch


def batch_create_video_youtube_bb_dataset_items(
    batch,
    dataset_id,
    dataset_version,
    project_id,
    rpc,
    input_type=None,
    num_dataset_items=0,
    dataset_item_ids=None,
):
    if not batch:
        logging.warning("Empty batch, nothing to process.")
        return batch

    logging.debug("Batch to create dataset items: %s", batch)

    for idx, video in enumerate(batch):
        total_frames = len(video)
        logging.debug("Total frames in video %d: %d", idx, total_frames)

        segment_length = 16
        logging.debug("Segment length for video %d: %d", idx, segment_length)

        # Ensure all frames share the same dataset_item_id
        dataset_item_id = video[0]["dataset_item_id"]
        for frame in video:
            if frame["dataset_item_id"] != dataset_item_id:
                raise ValueError(
                    f"Inconsistent dataset_item_id in video {idx}. Expected {dataset_item_id}, "
                    f"but got {frame['dataset_item_id']}"
                )

        segments = []
        current_segment = {}
        segment_id = 0

        for i, frame in enumerate(video):
            frame_num = frame["transformed_frame_number"]
            current_segment[frame_num] = {
                "fileLocation": frame["cloud_file_path"],
                "filename": frame["file_name"],
            }

            if len(current_segment) >= segment_length or i == total_frames - 1:
                segments.append(current_segment)
                current_segment = {}
                segment_id += 1


        logging.debug("Total segments created for video %d: %d", idx, len(segments))

        # STEP 2: Upload segments in chunks of 10
        for chunk_start in range(0, len(segments), 10):
            segment_chunk = segments[chunk_start:chunk_start + 10]
            frame_files_payload = {
                str(i): segment
                for i, segment in enumerate(segment_chunk, start=chunk_start)
            }

            payload = {
                "datasetId": dataset_id,
                "datasetItemId": dataset_item_id,
                "frameFiles": frame_files_payload,
            }
            
            logging.debug("Payload for uploading segments: %s", payload)

            logging.info("Uploading frame segments chunk for datasetItemId %s: %s",
                         dataset_item_id, list(frame_files_payload.keys()))

            resp = rpc.post(
                f"/v2/dataset/add_frame_segments?projectId={project_id}",
                payload=payload,
            )

            if not resp or not resp.get("success"):
                logging.warning("Failed to upload frame segments for chunk starting at segment %d", chunk_start)
            else:
                logging.info("Successfully uploaded segments %d–%d for datasetItemId %s",
                             chunk_start, chunk_start + len(segment_chunk) - 1, dataset_item_id)

    return batch


def batch_create_video_dataset_items(
    batch,
    dataset_id,
    dataset_version,
    project_id,
    rpc,
    input_type=None,
    num_dataset_items=0,
    dataset_item_ids=[]
):
    logging.debug("Creating batch_create_video_dataset_items with input_type %s", input_type)

    get_relative_path_fn = (
        get_davis_relative_path if input_type == "davis" else get_youtube_bb_relative_path
    )

    if not batch:
        logging.warning("Empty batch, nothing to process.")
        return batch

    logging.debug("Batch to create dataset items: %s", batch)

    segment_length = 16

    for idx, video in enumerate(batch):
        total_frames = len(video)
        logging.debug("Total frames in video %d: %d", idx, total_frames)

        dataset_item_id = video[0]["dataset_item_id"]

        # Validate dataset_item_id consistency
        for frame in video:
            if frame["dataset_item_id"] != dataset_item_id:
                raise ValueError(
                    f"Inconsistent dataset_item_id in video {idx}. "
                    f"Expected {dataset_item_id}, but got {frame['dataset_item_id']}"
                )

        segments = []
        current_segment = {}
        segment_id = 0

        for i, frame in enumerate(video):
            path = get_relative_path_fn(frame["local_file_path"])
            if path is None:
                path = frame["cloud_file_path"]

            if input_type == "mscoco_video":
                final_path= frame["cloud_file_path"]
            else:
                final_path = f"{dataset_id}/{path}"
            frame_num = frame["frame_number"]
            
            current_segment[frame_num] = {
                "fileLocation": final_path,
                "filename": frame["file_name"],
            }

            if len(current_segment) >= segment_length or i == total_frames - 1:
                segments.append(current_segment)
                current_segment = {}
                segment_id += 1

        logging.debug("Total segments created for video %d: %d", idx, len(segments))

        # Upload segments in chunks of 10
        for chunk_start in range(0, len(segments), 10):
            segment_chunk = segments[chunk_start:chunk_start + 10]
            frame_files_payload = {
                str(i): segment
                for i, segment in enumerate(segment_chunk, start=chunk_start)
            }

            payload = {
                "datasetId": dataset_id,
                "datasetItemId": dataset_item_id,
                "frameFiles": frame_files_payload,
            }

            logging.debug("Payload for uploading segments: %s", payload)
            logging.info("Uploading frame segments chunk for datasetItemId %s: %s",
                         dataset_item_id, list(frame_files_payload.keys()))

            resp = rpc.post(
                f"/v2/dataset/add_frame_segments?projectId={project_id}",
                payload=payload,
            )

            if not resp or not resp.get("success"):
                logging.warning("Failed to upload frame segments for chunk starting at segment %d", chunk_start)
            else:
                logging.info("Successfully uploaded segments %d–%d for datasetItemId %s",
                             chunk_start, chunk_start + len(segment_chunk) - 1, dataset_item_id)

    return batch


def batch_create_video_imagenet_dataset_items(
    batch,
    dataset_id,
    dataset_version,
    project_id,
    rpc,
    input_type=None,
):
    logging.debug(
        "creating batch_create_video_dataset_items with input_type %s",
        input_type,
    )
    get_relative_path_fn = (
        get_davis_relative_path
        if input_type in ["davis", "mscoco_video"]
        else get_youtube_bb_relative_path
    )
    if not batch:
        logging.warning("Empty batch, nothing to process.")
        return batch
    logging.debug("batch to create dataset items: %s", batch)
    video_files = []
    for video in batch:
        video_dict = {}
        path = get_relative_path_fn(video["local_file_path"])
        if path is None:
            path = video["cloud_file_path"]
            video_dict = {
                "fileLocation": f"{path}",
                "filename": video["file_name"],
            }
        else:
            video_dict = {
                "fileLocation": f"{dataset_id}/{path}",
                "filename": video["file_name"],
            }
        video_files.append(video_dict)
    payload = {
        "datasetId": dataset_id,
        "partitionNumber": batch[0]["partition_num"],
        "version": dataset_version,
        "videoFiles": video_files,
    }
    logging.info(
        "Payload for create dataset items: %s",
        payload,
    )
    resp = rpc.post(
        f"/v2/dataset/add_video_dataset_items?projectId={project_id}",
        payload=payload,
    )
    logging.info(
        "Response from create dataset items: %s",
        resp,
    )
    return batch


def batch_create_video_imagenet_items(
    batch,
    dataset_id,
    dataset_version,
    project_id,
    rpc,
    input_type=None,
):
    logging.debug(
        "creating batch_create_video_dataset_items with input_type %s",
        input_type,
    )
    get_relative_path_fn = (
        get_davis_relative_path if input_type == "davis" else get_youtube_bb_relative_path
    )
    if not batch:
        logging.warning("Empty batch, nothing to process.")
        return batch
    logging.debug("batch to create dataset items: %s", batch)
    frame_files = []
    for video in batch:
        video_dict = {}
        for frame in video:
            path = get_relative_path_fn(frame["local_file_path"])
            if path is None:
                path = frame["cloud_file_path"]
            video_dict[frame["frame_number"]] = {
                "fileLocation": f"{dataset_id}/{path}",
                "filename": frame["file_name"],
            }
        frame_files.append(video_dict)
    payload = {
        "datasetId": dataset_id,
        "partitionNumber": batch[0][0]["partition_num"],
        "version": dataset_version,
        "frameFiles": frame_files,
    }
    logging.info(
        "Payload for create dataset items: %s",
        payload,
    )
    resp = rpc.post(
        f"/v2/dataset/add_video_dataset_items?projectId={project_id}",
        payload=payload,
    )
    logging.info(
        "Response from create dataset items: %s",
        resp,
    )
    return batch


def get_client_annotations_processing_pipeline(
    annotations_partition,
    dataset_id,
    dataset_version,
    base_dataset_path,
    rpc,
    compress_annotations=False,
    max_attempts=5,
    batch_size=16,
    bucket_alias="",
    account_number="",
    project_id=None,
):
    logging.info("Setting up annotations pipeline")
    stage_1_queue = Queue()
    stage_2_queue = Queue()
    annotations_client_pipeline = Pipeline()
    if compress_annotations:
        annotations_partition["files"] = [
            compress_annotation_files(
                annotations_partition["files"],
                base_dataset_path,
            )
        ]
    batches = get_partition_batches(
        annotations_partition,
        batch_size,
        dataset_id,
        dataset_version,
        base_dataset_path,
        include_version_in_cloud_path=True,
    )
    for batch in batches:
        stage_1_queue.put(batch)
    update_annotation_bucket_url(
        rpc,
        dataset_id,
        batches[0][0]["partition_num"],
        annotation_bucket_url=os.path.commonpath(
            ["/".join(item["cloud_file_path"].split("/")[:-1]) for item in batches[0]]
        ).replace(os.sep, "/"),
    )  

    annotations_client_pipeline.add_stage(
        stage_name="fetching_presigned_urls",
        process_fn=add_batch_presigned_upload_urls,
        pull_queue=stage_1_queue,
        push_queue=stage_2_queue,
        process_params={
            "rpc": rpc,
            "partition_type": ANNOTATION_PARTITION_TYPE,
            "bucket_alias": bucket_alias,
            "account_number": account_number,
            "project_id": project_id,
        },
        num_threads=10,
    )
    annotations_client_pipeline.add_stage(
        stage_name="uploading_files",
        pull_queue=stage_2_queue,
        process_fn=upload_batch_files,
        process_params={"max_attempts": max_attempts},
        num_threads=10,
    )
    return annotations_client_pipeline


def get_client_images_processing_pipeline(
    images_partitions,
    dataset_id,
    dataset_version,
    base_dataset_path,
    rpc,
    max_attempts=5,
    batch_size=16,
    bucket_alias="",
    account_number="",
    project_id=None,
):
    logging.info("Setting up images pipeline")
    logging.debug(
        "images_partitions: %s, dataset_id: %s, dataset_version: %s, base_dataset_path: %s",
        images_partitions,
        dataset_id,
        dataset_version,
        base_dataset_path,
    )
    stage_1_queue = Queue()
    stage_2_queue = Queue()
    stage_3_queue = Queue()
    for partition in images_partitions:
        for batch in get_partition_batches(
            partition,
            batch_size,
            dataset_id,
            dataset_version,
            base_dataset_path,
        ):
            stage_1_queue.put(batch)
    images_client_pipeline = Pipeline()
    images_client_pipeline.add_stage(
        stage_name="fetching_presigned_urls",
        pull_queue=stage_1_queue,
        push_queue=stage_2_queue,
        process_fn=add_batch_presigned_upload_urls,
        process_params={
            "rpc": rpc,
            "partition_type": SAMPLES_PARTITION_TYPE,
            "bucket_alias": bucket_alias,
            "account_number": account_number,
            "project_id":project_id,
        },
        num_threads=10,
    )
    images_client_pipeline.add_stage(
        stage_name="uploading_files",
        pull_queue=stage_2_queue,
        push_queue=stage_3_queue,
        process_fn=upload_batch_files,
        process_params={"max_attempts": max_attempts},
        num_threads=10,
    )
    images_client_pipeline.add_stage(
        stage_name="inserting_dataset_items",
        pull_queue=stage_3_queue,
        process_fn=batch_create_dataset_items,
        process_params={
            "dataset_id": dataset_id,
            "dataset_version": dataset_version,
            "rpc": rpc,
        },
        num_threads=10,
    )
    return images_client_pipeline


def get_client_video_processing_pipeline(
    images_partitions,
    dataset_id,
    dataset_version,
    base_dataset_path,
    rpc,
    project_id,
    max_attempts=5,
    batch_size=16,
    bucket_alias="",
    account_number="",
    input_type="youtube_bb",
    num_dataset_items=0,
):
    logging.info("Setting up images pipeline")
    logging.debug(
        "images_partitions: %s, dataset_id: %s, dataset_version: %s, base_dataset_path: %s",
        images_partitions,
        dataset_id,
        dataset_version,
        base_dataset_path,
    )
    num_partitions = len(images_partitions)
    stage_1_queue = Queue()
    stage_2_queue = Queue()
    stage_3_queue = Queue()

    # Track which partition numbers we've already initialized
    initialized_partitions = set()

    for partition in images_partitions:
        partition_num = partition["partitionNum"]
        num_dataset_items = len(partition["files"])
        logging.debug(
            "num_dataset_items: %s, partition_num: %s",
            num_dataset_items,
            partition_num,
        )
        # Only create blank dataset items if this partition hasn't been processed
        if input_type in ["youtube_bb", "mot", "davis", "mscoco_video"]:
            if partition_num not in initialized_partitions:
                dataset_item_ids=create_video_blank_dataset_items(
                    partition_num,
                    rpc,
                    dataset_id,
                    dataset_version,
                    num_dataset_items,
                    project_id,
                )
                initialized_partitions.add(partition_num)

        if input_type == "youtube_bb":
            batches=get_video_youtube_bb_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
        elif input_type == "mot":
            batches= get_video_mot_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
        elif input_type == "davis":
            batches= get_video_davis_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
                
        elif input_type == "video_imagenet":
            batches=get_video_imagenet_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
               
        elif input_type == "kinetics":
            batches=get_video_kinetics_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
        elif input_type == "mscoco_video":
            batches= get_video_mscoco_partition_batches(
                partition,
                batch_size,
                dataset_id,
                dataset_version,
                base_dataset_path,
            )
        
        video_index = 0
        for batch in batches:
            if input_type in ["youtube_bb", "mot", "davis", "mscoco_video"]:
                for video in batch:
                    dataset_item_id = dataset_item_ids[video_index]
                    for frame in video:
                        frame["dataset_item_id"] = dataset_item_id
                    video_index += 1
            stage_1_queue.put(batch)


                
    images_client_pipeline = Pipeline()
    add_video_batch_presigned_upload_urls_fn = (
        add_video_imagenet_presigned_upload_urls
        if input_type
        in [
            "video_imagenet",
            "kinetics"
        ]
        else add_video_batch_presigned_upload_urls
    )
    upload_video_batch_files_fn = (
        upload_video_imagenet_batch_files
        if input_type
        in [
            "video_imagenet",
            "kinetics"
        ]
        else upload_video_batch_files
    )
    images_client_pipeline.add_stage(
        stage_name="fetching_presigned_urls",
        pull_queue=stage_1_queue,
        push_queue=stage_2_queue,
        process_fn=add_video_batch_presigned_upload_urls_fn,
        process_params={
            "rpc": rpc,
            "partition_type": SAMPLES_PARTITION_TYPE,
            "bucket_alias": bucket_alias,
            "account_number": account_number,
        },
        num_threads=10,
    )
    images_client_pipeline.add_stage(
        stage_name="uploading_files",
        pull_queue=stage_2_queue,
        push_queue=stage_3_queue,
        process_fn=upload_video_batch_files_fn,
        process_params={"max_attempts": max_attempts},
        num_threads=10,
    )
    if input_type == "youtube_bb":
        insert_dataset_items_process_fn = batch_create_video_youtube_bb_dataset_items
        
    elif input_type in [
        "video_imagenet",
        "kinetics"
    ]:
        insert_dataset_items_process_fn = batch_create_video_imagenet_dataset_items
    else:
        insert_dataset_items_process_fn = batch_create_video_dataset_items
    process_params = {
        "dataset_id": dataset_id,
        "dataset_version": dataset_version,
        "rpc": rpc,
        "project_id": project_id,
        "input_type": input_type,
    }
    if input_type in ["youtube_bb", "davis", "mscoco_video"]:
        process_params["num_dataset_items"] = num_dataset_items
        process_params["dataset_item_ids"] = dataset_item_ids
    images_client_pipeline.add_stage(
        stage_name="inserting_dataset_items",
        pull_queue=stage_3_queue,
        process_fn=insert_dataset_items_process_fn,
        process_params=process_params,
        num_threads=10,
    )
    return images_client_pipeline


def get_client_processing_pipelines(
    rpc,
    dataset_id,
    dataset_version,
    images_partition_status: list,
    annotation_partition_status: list,
    dataset_path: str,
    is_annotations_compressed: bool,
    destination_bucket_alias: str,
    account_number: str,
    project_id=None,
):
    annotation_pipeline, images_pipeline = (
        None,
        None,
    )
    if annotation_partition_status:
        logging.info("Setting up client annotations pipeline")
        annotation_pipeline = get_client_annotations_processing_pipeline(
            annotations_partition=annotation_partition_status,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            base_dataset_path=dataset_path,
            rpc=rpc,
            compress_annotations=is_annotations_compressed,
            bucket_alias=destination_bucket_alias,
            account_number=account_number,
            project_id=project_id,
        )
    if images_partition_status:
        logging.info("Setting up client images pipeline")
        images_pipeline = get_client_images_processing_pipeline(
            images_partitions=images_partition_status,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            base_dataset_path=dataset_path,
            rpc=rpc,
            bucket_alias=destination_bucket_alias,
            account_number=account_number,
            project_id=project_id,
        )
    return annotation_pipeline, images_pipeline


def get_client_video_processing_pipelines(
    project_id,
    rpc,
    dataset_id,
    dataset_version,
    images_partition_status: list,
    annotation_partition_status: list,
    dataset_path: str,
    is_annotations_compressed: bool,
    destination_bucket_alias: str,
    account_number: str,
    input_type: str = "youtube_bb",
    num_dataset_items: int = 0,
):
    logging.debug(
        "get_client_video_processing_pipelines with input_type %s",
        input_type,
    )
    annotation_pipeline, images_pipeline = (
        None,
        None,
    )
    if annotation_partition_status:
        annotation_pipeline = get_client_annotations_processing_pipeline(
            annotations_partition=annotation_partition_status,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            base_dataset_path=dataset_path,
            rpc=rpc,
            compress_annotations=is_annotations_compressed,
            bucket_alias=destination_bucket_alias,
            account_number=account_number,
            project_id=project_id,
        )
    if images_partition_status:
        images_pipeline = get_client_video_processing_pipeline(
            images_partitions=images_partition_status,
            dataset_id=dataset_id,
            dataset_version=dataset_version,
            base_dataset_path=dataset_path,
            rpc=rpc,
            project_id=project_id,
            bucket_alias=destination_bucket_alias,
            account_number=account_number,
            input_type=input_type,
            num_dataset_items=num_dataset_items,
        )
    return annotation_pipeline, images_pipeline


def handle_partition_stats(
    rpc,
    dataset_id,
    source_dataset_version,
    target_dataset_version,
    dataset_path,
    skip_annotation_pipeline,
):
    partition_stats = []
    (
        annotation_partition_status,
        images_partition_status,
    ) = get_partition_status(
        base_path=dataset_path,
        skip_annotation_partition=skip_annotation_pipeline,
    )
    logging.debug("Debug - annotation_partition_status: %s", annotation_partition_status)
    logging.debug("Debug - images_partition_status: %s", images_partition_status)
    if annotation_partition_status:
        partition_stats.append(annotation_partition_status)
    if images_partition_status:
        partition_stats.extend(images_partition_status)
    partition_stats = update_partitions_numbers(rpc, dataset_id, partition_stats)
    create_partition_stats(
        rpc,
        partition_stats,
        dataset_id,
        target_dataset_version,
        source_dataset_version,
    )
    return (
        annotation_partition_status,
        images_partition_status,
    )


def handle_video_partition_stats(
    rpc,
    dataset_id,
    source_dataset_version,
    target_dataset_version,
    dataset_path,
    skip_annotation_pipeline,
    input_type="youtube_bb",
):
    partition_stats = []
    rename_annotation_files = False
    if input_type == "youtube_bb":
        get_partitions_fn = get_youtube_bb_partitions
    elif input_type == "mot":
        get_partitions_fn = get_mot_partitions
        rename_annotation_files = True
    elif input_type == "davis":
        #restructure_davis_dataset(dataset_path)
        get_partitions_fn = get_davis_partitions
        rename_annotation_files = True
    elif input_type == "video_imagenet":
        get_partitions_fn = get_video_imagenet_partitions
    elif input_type == "kinetics":
        get_partitions_fn = get_kinetics_partitions
    elif input_type == "mscoco_video":
        ann_paths, im_paths = scan_dataset(dataset_path)
        paths = ann_paths + im_paths
        extract_frames_from_videos(paths)
        get_partitions_fn = get_video_mscoco_partitions
    (
        annotation_partition_status,
        images_partition_status,
        unique_videos
    ) = get_video_partition_status(
        base_path=dataset_path,
        skip_annotation_partition=skip_annotation_pipeline,
        get_partitions=get_partitions_fn,
        rename_annotation_files=rename_annotation_files,
        input_type=input_type.lower(),
    )
    if annotation_partition_status:
        partition_stats.append(annotation_partition_status)
    if images_partition_status:
        partition_stats.extend(images_partition_status)
    partition_stats = update_partitions_numbers(rpc, dataset_id, partition_stats)
    create_partition_stats(
        rpc,
        partition_stats,
        dataset_id,
        target_dataset_version,
        source_dataset_version,
    )
    return (
        annotation_partition_status,
        images_partition_status,
        unique_videos
    )


def start_client_processing_pipelines(
    rpc,
    dataset_id,
    dataset_version,
    images_partition_status,
    annotation_partition_status,
    dataset_path,
    is_annotations_compressed,
    destination_bucket_alias,
    account_number,
    project_id=None,
):
    annotation_pipeline, images_pipeline = get_client_processing_pipelines(
        rpc=rpc,
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        images_partition_status=images_partition_status,
        annotation_partition_status=annotation_partition_status,
        dataset_path=dataset_path,
        is_annotations_compressed=is_annotations_compressed,
        destination_bucket_alias=destination_bucket_alias,
        account_number=account_number,
        project_id=project_id,
    )
    if annotation_pipeline:
        logging.info("Starting annotation pipeline")
        annotation_pipeline.start()
        logging.info("Waiting for annotation pipeline to complete")
        annotation_pipeline.wait_to_finish_processing_and_stop()
    if images_pipeline:
        logging.info("Starting images pipeline")
        images_pipeline.start()
        logging.info("Waiting for images pipeline to complete")
        images_pipeline.wait_to_finish_processing_and_stop()


def start_client_video_processing_pipelines(
    project_id,
    rpc,
    dataset_id,
    dataset_version,
    images_partition_status,
    annotation_partition_status,
    dataset_path,
    is_annotations_compressed,
    destination_bucket_alias,
    account_number,
    input_type,
    num_dataset_items,
):
    annotation_pipeline, images_pipeline = get_client_video_processing_pipelines(
        project_id=project_id,
        rpc=rpc,
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        images_partition_status=images_partition_status,
        annotation_partition_status=annotation_partition_status,
        dataset_path=dataset_path,
        is_annotations_compressed=is_annotations_compressed,
        destination_bucket_alias=destination_bucket_alias,
        account_number=account_number,
        input_type=input_type.lower(),
        num_dataset_items=num_dataset_items,
    )
    if annotation_pipeline:
        logging.info("Starting annotation pipeline")
        annotation_pipeline.start()
        logging.info("Waiting for annotation pipeline to complete")
        annotation_pipeline.wait_to_finish_processing_and_stop()
    if images_pipeline:
        logging.info("Starting images pipeline")
        images_pipeline.start()
        logging.info("Waiting for images pipeline to complete")
        images_pipeline.wait_to_finish_processing_and_stop()


def handle_client_processing_pipelines(
    rpc,
    dataset_id,
    source_dataset_version,
    target_dataset_version,
    input_type,
    source_URL="",
    dataset_path="",
    destination_bucket_alias="",
    account_number="",
    skip_partition_status=False,
    annotation_partition_status=None,
    images_partition_status=None,
    project_id=None,
):
    is_annotations_compressed = input_type in [
        "pascalvoc",
        "pascal_voc",
        "yolo",
    ]
    skip_annotation_pipeline = (
        input_type in ["imagenet", "unlabeled"]
        or input_type == "labelbox"
        and not source_URL
        or input_type == "labelbox_classification"
        and not source_URL
    )
    logging.debug("Debug - skip_partition_status: %s", skip_partition_status)
    if not skip_partition_status:
        (
            annotation_partition_status,
            images_partition_status,
        ) = handle_partition_stats(
            rpc=rpc,
            dataset_id=dataset_id,
            source_dataset_version=source_dataset_version,
            target_dataset_version=target_dataset_version,
            dataset_path=dataset_path,
            skip_annotation_pipeline=skip_annotation_pipeline,
        )
    logging.info("Starting client processing pipelines for %s dataset", input_type)
    start_client_processing_pipelines(
        rpc=rpc,
        dataset_id=dataset_id,
        dataset_version=target_dataset_version,
        images_partition_status=images_partition_status,
        annotation_partition_status=annotation_partition_status,
        dataset_path=dataset_path,
        is_annotations_compressed=is_annotations_compressed,
        destination_bucket_alias=destination_bucket_alias,
        account_number=account_number,
        project_id=project_id,
    )


def handle_client_video_processing_pipelines(
    project_id,
    rpc,
    dataset_id,
    source_dataset_version,
    target_dataset_version,
    input_type,
    source_URL="",
    dataset_path="",
    destination_bucket_alias="",
    account_number="",
    skip_partition_status=False,
    annotation_partition_status=None,
    images_partition_status=None,
    unique_videos=0,
):
    is_annotations_compressed = input_type in [
        "davis",
        "mot",
    ]
    skip_annotation_pipeline = (
        input_type
        in [
            "imagenet",
            "unlabeled",
            "video_imagenet",
        ]
        or input_type == "labelbox"
        and not source_URL
        or input_type == "labelbox_classification"
        and not source_URL
    )
    if not skip_partition_status:
        (
            annotation_partition_status,
            images_partition_status,
            unique_videos,
        ) = handle_video_partition_stats(
            rpc=rpc,
            dataset_id=dataset_id,
            source_dataset_version=source_dataset_version,
            target_dataset_version=target_dataset_version,
            dataset_path=dataset_path,
            skip_annotation_pipeline=skip_annotation_pipeline,
            input_type=input_type.lower(),
        )
    start_client_video_processing_pipelines(
        project_id=project_id,
        rpc=rpc,
        dataset_id=dataset_id,
        dataset_version=target_dataset_version,
        images_partition_status=images_partition_status,
        annotation_partition_status=annotation_partition_status,
        dataset_path=dataset_path,
        is_annotations_compressed=is_annotations_compressed,
        destination_bucket_alias=destination_bucket_alias,
        account_number=account_number,
        input_type=input_type.lower(),
        num_dataset_items=unique_videos
    )
